from typing import Callable

import pytest

from antidote import Scope, world
from antidote._providers import IndirectProvider, ServiceProvider
from antidote.core.exceptions import DependencyNotFoundError
from antidote.exceptions import DuplicateDependencyError, FrozenWorldError


class Interface:
    pass


class A(Interface):
    pass


class B(Interface):
    pass


@pytest.fixture(autouse=True)
def empty_world():
    with world.test.empty():
        yield


@pytest.fixture
def service(empty_world):
    world.provider(ServiceProvider)
    return world.get(ServiceProvider)


@pytest.fixture
def indirect(empty_world):
    world.provider(IndirectProvider)
    return world.get(IndirectProvider)


@pytest.fixture(params=[True, False])
def permanent(request):
    return request.param


def test_implementation(permanent: bool):
    indirect = IndirectProvider()
    world.test.singleton(A, A())
    impl = indirect.register_implementation(Interface, lambda: A, permanent=permanent)

    assert world.test.maybe_provide_from(indirect, impl).unwrapped is world.get(A)
    assert world.test.maybe_provide_from(indirect, impl).is_singleton() is permanent
    assert Interface.__name__ in repr(indirect)


@pytest.mark.parametrize('singleton', [True, False])
@pytest.mark.parametrize('permanent', [True, False])
def test_implementation_permanent_singleton(service: ServiceProvider,
                                            singleton: bool,
                                            permanent: bool):
    scope = Scope.singleton() if singleton else None
    choice = 'a'

    def implementation():
        return dict(a=A, b=B)[choice]

    service.register(A, scope=scope)
    service.register(B, scope=scope)
    indirect = IndirectProvider()
    impl = indirect.register_implementation(Interface,
                                            implementation,
                                            permanent=permanent)

    instance = world.test.maybe_provide_from(indirect, impl).unwrapped
    assert isinstance(instance, A)
    assert (instance is world.get(A)) is singleton
    assert world.test.maybe_provide_from(
        indirect,
        impl).is_singleton() is (singleton and permanent)

    choice = 'b'
    assert implementation() == B
    assert world.test.maybe_provide_from(
        indirect,
        impl).is_singleton() is (singleton and permanent)
    instance = world.test.maybe_provide_from(indirect, impl).unwrapped
    if permanent:
        assert isinstance(instance, A)
        assert (instance is world.get(A)) is singleton
    else:
        assert isinstance(instance, B)
        assert (instance is world.get(B)) is singleton


def test_implementation_exists(indirect: IndirectProvider, permanent: bool):
    world.test.singleton(A, A())
    impl = indirect.register_implementation(Interface, lambda: A, permanent=permanent)

    assert not indirect.exists(object())
    assert indirect.exists(impl)
    assert not indirect.exists(A)


@pytest.mark.parametrize('keep_singletons_cache', [True, False])
@pytest.mark.parametrize('register', [
    pytest.param(
        lambda indirect, inf, impl: indirect.register_implementation(inf, lambda: impl,
                                                                     permanent=False),
        id='implementation'),
    pytest.param(
        lambda indirect, inf, impl: indirect.register_implementation(inf, lambda: impl,
                                                                     permanent=True),
        id='implementation_permanent')
])
def test_clone(indirect: IndirectProvider,
               keep_singletons_cache: bool,
               register: Callable[[IndirectProvider, type, type], object]):
    world.test.singleton({A: A()})

    dep = register(indirect, Interface, A)
    a = world.get(dep)
    assert isinstance(a, Interface)

    if keep_singletons_cache:
        with world.test.clone(keep_singletons=True):
            clone = indirect.clone(True)
            assert world.test.maybe_provide_from(clone, dep).unwrapped is a
    else:
        with world.test.empty():
            world.test.singleton({A: A(), B: B()})
            clone = indirect.clone(False)
            instance = world.test.maybe_provide_from(clone, dep).unwrapped
            assert instance is world.get(A)
            assert instance is not a

    class Interface2:
        pass

    class A2(Interface2):
        pass

    class Interface3:
        pass

    class A3(Interface3):
        pass

    world.test.singleton({A2: A2(), A3: A3()})
    # Original does not modify clone
    impl = register(indirect, Interface2, A2)
    assert world.get(impl) is world.get(A2)
    assert world.test.maybe_provide_from(clone, impl) is None
    assert clone.maybe_debug(impl) is None

    # Did not modify original provider
    impl = register(clone, Interface3, A3)
    assert world.test.maybe_provide_from(clone, impl).unwrapped is world.get(A3)
    assert indirect.maybe_debug(impl) is None
    with pytest.raises(DependencyNotFoundError):
        world.get(Interface3)


def test_freeze(indirect: IndirectProvider, permanent):
    world.freeze()

    with pytest.raises(FrozenWorldError):
        indirect.register_implementation(Interface, lambda: A, permanent=False)

    with pytest.raises(FrozenWorldError):
        indirect.register_implementation(Interface, lambda: A, permanent=True)

    with pytest.raises(DependencyNotFoundError):
        world.get(Interface)


def test_register_duplicate_check(indirect: IndirectProvider, permanent: bool):
    def implementation():
        return A

    indirect.register_implementation(Interface, implementation, permanent=permanent)

    with pytest.raises(DuplicateDependencyError):
        indirect.register_implementation(Interface, implementation, permanent=False)

    with pytest.raises(DuplicateDependencyError):
        indirect.register_implementation(Interface, implementation, permanent=True)


def test_invalid_link(permanent: bool):
    indirect = IndirectProvider()
    target = 'target'

    def implementation():
        return target

    impl = indirect.register_implementation(A, implementation, permanent=permanent)

    with pytest.raises(DependencyNotFoundError, match=".*" + target + ".*"):
        world.test.maybe_provide_from(indirect, impl)



# pyright: reportUnnecessaryIsInstance=false
import functools
import inspect
import warnings
from typing import Callable, cast, Type, TypeVar

from typing_extensions import ParamSpec, Protocol

from ._implementation import ImplementationWrapper, validate_provided_class
from ._internal import API
from ._internal.wrapper import is_wrapper
from ._providers import IndirectProvider
from .core import inject
from .core.exceptions import DoubleInjectionError

P = ParamSpec('P')
T = TypeVar('T')


@API.private
class ImplementationProtocol(Protocol[P, T]):
    """
    :meta private:
    """

    def __rmatmul__(self, klass: type) -> object:  # pragma: no cover
        warnings.warn("Use the new `world.get(<dependency>, source=<implementation>)` syntax.")
        return  # type: ignore

    def __antidote_dependency__(self, target: Type[T]) -> object:
        ...  # pragma: no cover

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        ...  # pragma: no cover


@API.public
def implementation(interface: type,
                   *,
                   permanent: bool = True
                   ) -> Callable[[Callable[P, T]], ImplementationProtocol[P, T]]:
    """
    Defines which dependency should be retrieved when :code:`interface` is requested.
    Suppose we have to support two different databases and the configuration defines
    which one should be chosen:

    .. doctest:: implementation

        >>> from antidote import Constants, const
        >>> class Config(Constants):
        ...     DB = const('postgres')
        >>> # The interface
        ... class Database:
        ...     pass

    .. doctest:: implementation

        >>> from antidote import Service, factory
        >>> class PostgreSQL(Database, Service):
        ...     pass
        >>> class MySQL(Database):
        ...     pass
        >>> @factory
        ... def build_mysql() -> MySQL:
        ...     return MySQL()

    The decorated function must return the dependency that should be used for
    :code:`Database`. It'll be automatically injected, so you can use annotated type hints
    on it. Furthermore, the chosen implementation is by default permanent. Meaning that
    the decorated function will only be called once, at most.

    .. doctest:: implementation

        >>> from antidote import implementation, Get
        >>> from typing import Annotated
        ... # from typing_extensions import Annotated # Python < 3.9
        >>> @implementation(Database)
        ... def local_db(choice: Annotated[str, Get(Config.DB)]) -> object:
        ...     if choice == 'postgres':
        ...         return PostgreSQL
        ...     else:
        ...         return MySQL @ build_mysql

    To retrieve the actual implementation from Antidote you need to use a specific syntax
    :code:`dependency @ implementation` as presented in the following examples.
    The goal of it is twofold:

    - Ensure that the implementation is loaded whenever you require the interface.
    - Better maintainability as you know *where* the dependenciy comes from.

    .. doctest:: implementation

        >>> from antidote import world, inject
        >>> world.get(Database @ local_db)  # treated as `object` by Mypy
        <PostgreSQL ...>
        >>> # With Mypy casting
        ... world.get[Database](Database @ local_db)
        <PostgreSQL ...>
        >>> # Concise Mypy casting
        ... world.get[Database] @ local_db
        <PostgreSQL ...>
        >>> @inject([Database @ local_db])
        ... def f(db: Database):
        ...     pass

    Or with annotated type hints:

    .. doctest:: implementation

        >>> from typing import Annotated
        ... # from typing_extensions import Annotated # Python < 3.9
        >>> from antidote import From
        >>> @inject
        ... def f(db: Annotated[Database, From(local_db)]):
        ...     pass

    Args:
        interface: Interface for which an implementation will be provided
        permanent: Whether the function should be called each time the interface is needed
            or not. Defaults to :py:obj:`True`.

    Returns:
        The decorated function, unmodified.
    """
    if not isinstance(permanent, bool):
        raise TypeError(f"permanent must be a bool, not {type(permanent)}")
    if not inspect.isclass(interface):
        raise TypeError(f"interface must be a class, not {type(interface)}")

    @inject
    def register(func: Callable[P, T],
                 indirect_provider: IndirectProvider = inject.me()
                 ) -> ImplementationProtocol[P, T]:
        if not (inspect.isfunction(func)
                or (is_wrapper(func)
                    and inspect.isfunction(func.__wrapped__))):  # type: ignore
            raise TypeError(f"{func} is not a function")

        try:
            func = inject(func)
        except DoubleInjectionError:
            pass

        @functools.wraps(func)
        def impl() -> object:
            dep = func()
            validate_provided_class(dep, expected=interface)
            return dep

        dependency = indirect_provider.register_implementation(interface, impl,
                                                               permanent=permanent)
        return ImplementationWrapper[P, T](func, dependency)

    return cast(Callable[[Callable[P, T]], ImplementationProtocol[P, T]], register)



import pytest

from antidote import LazyCall, LazyMethodCall, Service, world
from antidote._providers import LazyProvider, ServiceProvider


@pytest.fixture(autouse=True)
def empty_world():
    with world.test.empty():
        world.provider(LazyProvider)
        world.provider(ServiceProvider)
        yield


def test_lazy_call_singleton():
    def func():
        return object()

    test = LazyCall(func, singleton=False)
    assert world.get(test) != world.get(test)

    test2 = LazyCall(func, singleton=True)
    assert world.get(test2) is world.get(test2)

    default = LazyCall(func)
    assert world.get(default) is world.get(default)


@pytest.mark.parametrize(
    'args,kwargs',
    [
        ((1,), dict(test=2)),
        ((), dict(x=32)),
        ((4,), {})
    ]
)
def test_args_kwargs(args, kwargs):
    def func(*args_, **kwargs_):
        return args_, kwargs_

    test = LazyCall(func)(*args, **kwargs)
    assert (args, kwargs) == world.get(test)


def test_method_call():
    x = object()

    class Test(Service):
        def get(self, s):
            return s

        A = LazyMethodCall(get)(x)

    assert world.get(Test.A) is x


def test_method_same_instance():
    class Test(Service):
        def get(self):
            return self

        A = LazyMethodCall(get, singleton=True)
        B = LazyMethodCall(get, singleton=False)

    instance = world.get(Test)
    assert world.get(Test.A) is instance
    assert world.get(Test.B) is instance


def test_method_singleton():
    class Test(Service):
        def get(self):
            return object()

        A = LazyMethodCall(get, singleton=True)
        B = LazyMethodCall(get, singleton=False)
        C = LazyMethodCall(get)

    assert world.get(Test.A) is world.get(Test.A)
    assert world.get(Test.B) != world.get(Test.B)
    assert world.get(Test.C) is world.get(Test.C)  # singleton by default


def test_method_direct_call():
    class Test(Service):
        def get(self):
            return "Hello"

        A = LazyMethodCall(get)

    t = Test()
    assert t.A == "Hello"


@pytest.mark.parametrize(
    'args,kwargs',
    [
        ((1,), dict(test=2)),
        ((), dict(x=32)),
        ((4,), {})
    ]
)
def test_method_args_kwargs(args, kwargs):
    class Test(Service):
        def get(self, *args_, **kwargs_):
            return args_, kwargs_

        A = LazyMethodCall(get)(*args, **kwargs)

    assert (args, kwargs) == Test().A
    assert (args, kwargs) == world.get(Test.A)


def test_invalid_lazy_call():
    with pytest.raises(TypeError, match=".*func.*"):
        LazyCall(func=object())

    with pytest.raises(TypeError, match=".*singleton.*"):
        LazyCall(func=lambda: None, singleton=object())


def test_invalid_lazy_method_call():
    with pytest.raises(TypeError, match=".*method.*"):
        LazyMethodCall(method=object())

    with pytest.raises(TypeError, match=".*singleton.*"):
        LazyMethodCall(method=lambda: None, singleton=object())


def test_invalid_lazy_method_class():
    class Dummy:
        def get(self, x):
            pass

        A = LazyMethodCall(get)
        B = LazyMethodCall(get)(x=1)

    with pytest.raises(RuntimeError):
        Dummy.A

    with pytest.raises(RuntimeError):
        Dummy().A

    with pytest.raises(RuntimeError):
        Dummy.B

    with pytest.raises(RuntimeError):
        Dummy().B



import pytest

from antidote import world
from antidote._providers import WorldTestProvider
from antidote.core import DependencyValue, Scope
from antidote.exceptions import (DuplicateDependencyError)


class Service:
    pass


class AnotherService:
    pass


@pytest.fixture()
def provider():
    with world.test.empty():
        yield world.get(WorldTestProvider)


def test_singleton(provider: WorldTestProvider):
    s = Service()
    assert not provider.exists(Service)

    provider.add_singletons({Service: s})
    assert world.get(Service) is s
    assert provider.exists(Service)


def test_factory(provider: WorldTestProvider):
    s = Service()

    assert not provider.exists(Service)
    provider.add_factory(Service, factory=lambda: s, scope=None)

    assert world.get(Service) is s
    assert provider.exists(Service)


def test_factory_singleton(provider: WorldTestProvider):
    provider.add_factory(Service, factory=lambda: Service(), scope=Scope.singleton())
    assert world.get(Service) is world.get(Service)


def test_factory_no_scope(provider: WorldTestProvider):
    provider.add_factory(Service, factory=lambda: Service(), scope=None)
    assert world.get(Service) is not world.get(Service)


def test_factory_scope(provider: WorldTestProvider):
    scope = world.scopes.new(name='dummy')
    provider.add_factory(Service, factory=lambda: Service(), scope=scope)

    s = world.get(Service)
    assert world.get(Service) is s

    world.scopes.reset(scope)
    assert world.get(Service) is not s


def test_duplicate():
    with world.test.empty():
        provider = world.get(WorldTestProvider)
        provider.add_singletons({Service: Service()})

        with pytest.raises(DuplicateDependencyError):
            provider.add_singletons({Service: Service()})

        with pytest.raises(DuplicateDependencyError):
            provider.add_factory(Service, factory=lambda: Service(), scope=None)

    with world.test.new():
        provider = world.get(WorldTestProvider)
        provider.add_factory(Service, factory=lambda: Service(), scope=None)

        with pytest.raises(DuplicateDependencyError):
            provider.add_singletons({Service: Service()})

        with pytest.raises(DuplicateDependencyError):
            provider.add_factory(Service, factory=lambda: Service(), scope=None)


def test_clone(provider: WorldTestProvider):
    s = Service()
    a_s = AnotherService()
    provider.add_singletons({Service: s})
    provider.add_factory(AnotherService, factory=lambda: a_s, scope=Scope.singleton())

    def d(x):
        return DependencyValue(x, scope=Scope.singleton())

    clone = provider.clone(keep_singletons_cache=True)
    assert world.test.maybe_provide_from(clone, Service) == d(s)
    assert world.test.maybe_provide_from(clone, AnotherService) == d(a_s)

    clone_without_singletons = provider.clone(keep_singletons_cache=False)
    assert world.test.maybe_provide_from(clone_without_singletons, Service) is None
    assert world.test.maybe_provide_from(clone_without_singletons,
                                         AnotherService) == d(a_s)



# @formatter:off
from cpython.ref cimport PyObject

from antidote._internal.stack cimport DependencyStack
# @formatter:on

ctypedef unsigned short Header
ctypedef unsigned short ScopeId

cdef bint header_is_singleton(Header header)
cdef Header header_scope(ScopeId scope_id)
cdef Header header_flag_singleton()
cdef Header header_flag_no_scope()
cdef Header header_flag_cacheable()

cdef class HeaderObject:
    cdef:
        Header header

    @staticmethod
    cdef HeaderObject from_scope(Scope scope)

cdef struct DependencyResult:
    # Holds information on the dependency (singleton, etc...).
    # Only valid if value is not NULL.
    Header header
    # Pointer to dependency value. Once done using the result, you MUST use Py_XDECREF
    # NULL if not found.
    PyObject *value

cdef class Scope:
    cdef:
        readonly str name
        ScopeId id

cdef class DependencyValue:
    cdef:
        readonly object unwrapped
        readonly Scope scope

    cdef to_result(self, DependencyResult *result)

    @staticmethod
    cdef DependencyValue from_result(RawContainer container, DependencyResult *result)

cdef class Container:
    pass

cdef class RawProvider:
    cdef:
        object _container_ref

    cdef object provider_type(self)
    cpdef RawProvider clone(self, bint keep_singletons_cache)
    cdef fast_provide(self,
                      PyObject *dependency,
                      PyObject *container,
                      DependencyResult *result)
    cdef RawContainer _bound_container(self)

cdef class FastProvider(RawProvider):
    pass

cdef class RawContainer(Container):
    cdef:
        object __weakref__

        DependencyStack _dependency_stack
        object _instantiation_lock
        object _registration_lock

        bint __frozen
        dict __singletons
        list __providers
        list __scopes
        list __scope_dependencies

        unsigned long __singletons_clock
        object __cache

    cdef Scope get_scope(self, ScopeId scope_id)
    cdef fast_get(self, PyObject *dependency, DependencyResult *result)
    cdef __safe_cache_provide(self,
                              PyObject *dependency,
                              DependencyResult *result,
                              CacheValue*cached)
    cdef __safe_provide(self,
                        PyObject *dependency,
                        DependencyResult *result,
                        unsigned long singletons_clock)

cdef struct CacheValue:
    Header header
    PyObject *ptr



from __future__ import annotations

import base64
import inspect
import textwrap
from collections import deque
from typing import (Deque, Hashable, List, Optional, Sequence, Set, Tuple, TYPE_CHECKING)

from .immutable import Immutable
from .. import API

if TYPE_CHECKING:
    from ...core.container import RawContainer, Scope

# Object will be allocated on the heap, so as close as possible to most user objects
# in memory.
_ID_MASK = id(object())


@API.private
def short_id(__obj: object) -> str:
    """ Produces a short, human readable, representation of the id of an object. """
    n = id(__obj) ^ _ID_MASK
    return (base64
            .b64encode(n.to_bytes(8, byteorder='little'))
            .decode('ascii')
            .rstrip('=')  # Remove padding
            .rstrip('A'))  # Remove 000000


@API.private
def debug_repr(__obj: object) -> str:
    from ..wrapper import is_wrapper
    try:
        return str(__obj.__antidote_debug_repr__())  # type: ignore
    except Exception:
        pass
    if (isinstance(__obj, type) and inspect.isclass(__obj)) \
            or inspect.isfunction(__obj) \
            or is_wrapper(__obj):
        if isinstance(__obj.__module__, str) \
                and __obj.__module__ not in {"__main__", "builtins"}:
            module = __obj.__module__ + "."
        else:
            module = ""
        return f"{module}{__obj.__qualname__}"  # type: ignore
    return repr(__obj)


@API.private
def get_injections(__func: object) -> Sequence[object]:
    from ..wrapper import get_wrapper_injections
    try:
        return list(get_wrapper_injections(__func).values())  # type: ignore
    except TypeError:
        return []


@API.private
class Task(Immutable):
    __slots__ = ()


@API.private
class DependencyTask(Task):
    __slots__ = ('dependency',)
    dependency: Hashable


@API.private
class InjectionTask(Task):
    __slots__ = ('name', 'injections')
    name: str
    injections: List[Hashable]


@API.private
def scope_repr(scope: 'Optional[Scope]', *, empty: str) -> str:
    from ...core import Scope
    if scope is None:
        return "<∅> "
    elif scope is Scope.singleton() or scope is Scope.sentinel():
        return empty
    else:
        return f"<{scope.name}> "


_LEGEND = """
Singletons have no scope markers.
<∅> = no scope (new instance each time)
<name> = custom scope
"""


@API.private
def tree_debug_info(container: 'RawContainer',
                    origin: object,
                    max_depth: int = -1) -> str:
    from ...core.wiring import WithWiringMixin
    from ...core.exceptions import DependencyNotFoundError
    from ...core import Scope

    @API.private
    class DebugTreeNode(Immutable):
        __slots__ = ('info', 'scope', 'children', 'depth')
        info: str
        scope: Optional[Scope]
        children: List[DebugTreeNode]
        depth: int

        def __init__(self,
                     info: str,
                     *,
                     depth: int = 0,
                     scope: Optional[Scope] = Scope.sentinel(),
                     children: Optional[List[DebugTreeNode]] = None) -> None:
            super().__init__(textwrap.dedent(info),
                             scope,
                             children or [],
                             depth)

        def child(self,
                  info: str,
                  *,
                  scope: Optional[Scope] = Scope.sentinel()
                  ) -> DebugTreeNode:
            self.children.append(DebugTreeNode(
                info,
                depth=self.depth + 1,
                scope=scope,
            ))
            return self.children[-1]

    if max_depth < 0:
        max_depth = 1 << 31  # roughly infinity in this case.

    root = DebugTreeNode(debug_repr(origin))
    original_root = root
    tasks: Deque[Tuple[DebugTreeNode, Set[object], Task]] = deque([
        (root, set(), DependencyTask(origin))
    ])

    def add_root_injections(parent: DebugTreeNode,
                            parent_dependencies: Set[object],
                            dependency: Hashable) -> None:
        from ...core.wiring import Methods

        if isinstance(dependency, type) and inspect.isclass(dependency):
            cls = dependency
            conf = getattr(cls, '__antidote__', None)
            if conf is not None \
                    and isinstance(conf, WithWiringMixin) \
                    and conf.wiring is not None:
                if isinstance(conf.wiring.methods, Methods):
                    for name, member in cls.__dict__.items():
                        if name != '__init__' and callable(member):
                            injections = get_injections(member)
                            if injections:
                                tasks.append((parent, parent_dependencies, InjectionTask(
                                    name=f"Method: {name}",
                                    injections=injections,
                                )))
                else:
                    for name in sorted(conf.wiring.methods):
                        if name != '__init__':
                            tasks.append((parent, parent_dependencies, InjectionTask(
                                name=f"Method: {name}",
                                injections=get_injections(getattr(cls, name)),
                            )))
        elif callable(dependency):
            for d in get_injections(dependency):
                tasks.append((parent, parent_dependencies, DependencyTask(d)))

    while tasks:
        parent, parent_dependencies, task = tasks.pop()
        if isinstance(task, DependencyTask):
            dependency = task.dependency
            try:
                debug = container.debug(dependency)
            except DependencyNotFoundError:
                if dependency is origin:
                    add_root_injections(parent, parent_dependencies, dependency)
                else:
                    parent.child(f"/!\\ Unknown: {debug_repr(dependency)}")

                continue

            if dependency in parent_dependencies:
                parent.child(f"/!\\ Cyclic dependency: {debug.info}")
                continue

            child = parent.child(debug.info, scope=debug.scope)
            child_dependencies = parent_dependencies | {dependency}

            if dependency is origin:
                root = child  # previous root is redundant
                add_root_injections(child, child_dependencies, dependency)

            if child.depth < max_depth:
                for d in debug.dependencies:
                    tasks.append((child, child_dependencies,
                                  DependencyTask(d)))
                for w in debug.wired:
                    if isinstance(w, type) and inspect.isclass(w):
                        for d in get_injections(getattr(w, '__init__')):
                            tasks.append((child, child_dependencies,
                                          DependencyTask(d)))
                    else:
                        tasks.append((child, child_dependencies, InjectionTask(
                            name=debug_repr(w),
                            injections=get_injections(w),
                        )))
        elif isinstance(task, InjectionTask) and task.injections:
            child = parent.child(task.name)
            for d in task.injections:
                tasks.append((child, parent_dependencies, DependencyTask(d)))

    if not root.children and original_root is root:
        return f"{origin!r} is neither a dependency nor is anything injected."

    output = [
        scope_repr(root.scope, empty="") + root.info
    ]
    nodes: Deque[Tuple[str, bool, DebugTreeNode]] = deque([
        ("", i == 0, child)
        for i, child in enumerate(root.children[::-1])
    ])

    while nodes:
        prefix, last, node = nodes.pop()
        first_line, *rest = node.info.split("\n", 1)
        txt = prefix + ("└──" if last else "├──")
        txt += scope_repr(node.scope, empty=" ") + first_line
        new_prefix = prefix + ("    " if last else "│   ")
        if rest:
            txt += "\n" + textwrap.indent(rest[0], new_prefix)
        output.append(txt)

        for i, child in enumerate(node.children[::-1]):
            nodes.append((new_prefix, i == 0, child))

    output.append(_LEGEND)

    return "\n".join(output)



from typing import Callable

import pytest
from typing_extensions import Annotated

from antidote import factory, From, FromArg, Get, Service, world
from antidote._internal.world import LazyDependency
from antidote._providers import FactoryProvider, ServiceProvider
from antidote.exceptions import DependencyNotFoundError, FrozenWorldError
from .utils import DummyIntProvider


@pytest.fixture(autouse=True)
def empty_world():
    with world.test.empty():
        yield


class A:
    pass


def test_get():
    world.provider(ServiceProvider)
    a = A()
    world.test.singleton(A, a)
    assert world.get(A) is a

    class B(Service):
        pass

    assert world.get[A](A) is a
    assert world.get[B]() is world.get(B)

    b = B()
    assert world.get(B, default=b) is world.get(B)
    assert world.get[B](default=b) is world.get(B)

    with pytest.raises(DependencyNotFoundError):
        world.get(Service)

    with pytest.raises(DependencyNotFoundError):
        world.get[Service]()

    assert world.get[Service](default=b) is b


def test_get_type_safety():
    x = object()
    world.test.singleton(A, x)

    assert world.get(A) is x  # without type hints, it should not fail
    with pytest.raises(TypeError):
        world.get[A]()

    class B:
        pass

    assert world.get(B, default=x) is x
    with pytest.raises(TypeError, match="(?i).*default.*"):
        world.get[B](default=x)


def test_get_factory():
    world.provider(FactoryProvider)

    @factory
    def build_a() -> A:
        return A()

    assert isinstance(world.get(A, source=build_a), A)
    assert world.get[A](source=build_a) is world.get(A, source=build_a)
    assert world.get[A] @ build_a is world.get(A, source=build_a)
    assert world.get(A @ build_a) is world.get(A, source=build_a)


@pytest.mark.parametrize('getter', [
    pytest.param(world.get, id='get'),
    pytest.param(world.get[A], id='get[A]'),
    pytest.param(lambda x: world.lazy(x).get(), id='lazy'),
    pytest.param(lambda x: world.lazy[A](x).get(), id='lazy[A]')
])
def test_annotation_support(getter: Callable[[object], object]):
    class Maker:
        def __rmatmul__(self, other):
            return 'maker'

    world.test.singleton({
        A: A(),
        'a': A(),
        'maker': A()
    })
    assert getter(Annotated[A, object()]) is world.get(A)
    assert getter(Annotated[A, Get('a')]) is world.get('a')  # noqa: F821
    assert getter(Annotated[A, From(Maker())]) is world.get('maker')

    with pytest.raises(TypeError):
        getter(Annotated[A, Get('a'), Get('a')])  # noqa: F821

    with pytest.raises(TypeError):
        getter(Annotated[A, FromArg(lambda a: a)])  # noqa: F821


def test_lazy():
    world.test.singleton({
        'x': object(),
        A: A()
    })

    lazy = world.lazy('x')
    assert isinstance(lazy, LazyDependency)
    assert lazy.unwrapped == 'x'
    assert lazy.get() == world.get('x')

    lazy = world.lazy[A](A)
    assert isinstance(lazy, LazyDependency)
    assert lazy.unwrapped == A
    assert lazy.get() == world.get(A)

    assert world.lazy[A]().get() is world.get(A)


def test_lazy_type_safety():
    x = object()
    world.test.singleton(A, x)

    assert world.lazy(A).get() is x

    with pytest.raises(TypeError):
        world.lazy[A]().get()


def test_lazy_factory():
    world.provider(FactoryProvider)

    @factory
    def build_a() -> A:
        return A()

    assert (world.lazy[A] @ build_a).get() is world.get(A @ build_a)


def test_freeze():
    world.provider(ServiceProvider)
    provider = world.get[ServiceProvider]()

    class Service:
        pass

    world.freeze()
    with pytest.raises(FrozenWorldError):
        world.test.singleton("test", "x")

    with pytest.raises(FrozenWorldError):
        provider.register(Service, scope=None)


def test_add_provider():
    world.provider(DummyIntProvider)
    assert world.get(10) == 20


def test_no_duplicate_provider():
    world.provider(DummyIntProvider)
    assert world.get(10) == 20

    with pytest.raises(ValueError, match=".*already exists.*"):
        world.provider(DummyIntProvider)


@pytest.mark.parametrize('p, expectation', [
    (object(), pytest.raises(TypeError, match=".*RawProvider.*")),
    (A, pytest.raises(TypeError, match=".*RawProvider.*"))
])
def test_invalid_add_provider(p, expectation):
    with expectation:
        world.provider(p)



import pytest

from antidote import Scope, world
from antidote._providers.service import Parameterized, ServiceProvider
from antidote.exceptions import (DependencyNotFoundError, DuplicateDependencyError,
                                 FrozenWorldError)


@pytest.fixture
def provider():
    with world.test.empty():
        world.provider(ServiceProvider)
        yield world.get(ServiceProvider)


@pytest.fixture(params=[
    pytest.param(None, id='scope=None'),
    pytest.param(Scope.singleton(), id='scope=singleton')
])
def scope(request):
    return request.param


class KeepInit:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class A(KeepInit):
    pass


class B(KeepInit):
    pass


@pytest.mark.parametrize(
    'wrapped,kwargs',
    [
        (1, {'test': 1}),
        (A, {'another': 'no'}),
        (A, {'not_hashable': {'hey': 'hey'}})
    ]
)
def test_parameterized_str_repr(wrapped, kwargs):
    b = Parameterized(wrapped, kwargs)

    # does not fail
    hash(b)

    assert repr(wrapped) in repr(b)
    assert repr(kwargs) in repr(b)


def test_parameterized_eq_hash():
    a = Parameterized(A, dict(test=1))
    assert hash(a) == hash(a)
    assert a == a

    a2 = Parameterized(A, dict(test=1))
    assert hash(a) == hash(a2)
    assert a == a2

    for x in [Parameterized(A, dict(test=2)),
              Parameterized(A, dict(hey=1)),
              Parameterized(1, dict(test=1))]:
        assert hash(a) != hash(x)
        assert a != x


def test_simple(provider: ServiceProvider, scope: Scope):
    provider.register(A, scope=scope)
    assert isinstance(world.get(A), A)
    assert repr(A) in repr(provider)


@pytest.mark.parametrize('singleton', [True, False])
def test_register(singleton: bool):
    with world.test.empty():
        provider = ServiceProvider()
        provider.register(A, scope=Scope.singleton() if singleton else None)
        assert world.test.maybe_provide_from(provider, A).is_singleton() is singleton
        assert isinstance(world.test.maybe_provide_from(provider, A).unwrapped, A)


def test_parameterized(scope: Scope):
    provider = ServiceProvider()
    provider.register(A, scope=scope)

    s = world.test.maybe_provide_from(provider,
                                      Parameterized(A, dict(val=object))).unwrapped
    assert isinstance(s, A)
    assert dict(val=object) == s.kwargs


def test_duplicate_error(provider: ServiceProvider, scope: Scope):
    provider.register(A, scope=scope)

    with pytest.raises(DuplicateDependencyError):
        provider.register(A, scope=scope)


@pytest.mark.parametrize('keep_singletons_cache', [True, False])
def test_copy(provider: ServiceProvider,
              keep_singletons_cache: bool, scope: Scope):
    class C:
        pass

    world.test.singleton('factory', lambda: C())
    provider.register(A, scope=scope)

    cloned = provider.clone(keep_singletons_cache)
    if keep_singletons_cache:
        with world.test.clone(keep_singletons=True):
            assert isinstance(world.test.maybe_provide_from(cloned, A).unwrapped, A)
    else:
        with world.test.clone(keep_singletons=False):
            assert isinstance(world.test.maybe_provide_from(cloned, A).unwrapped, A)

    class D:
        pass

    class E:
        pass

    # changing original does not change cloned
    provider.register(D, scope=scope)
    assert isinstance(world.get(D), D)
    assert world.test.maybe_provide_from(cloned, D) is None

    # changing cloned does not change original
    cloned.register(E, scope=scope)
    assert isinstance(world.test.maybe_provide_from(cloned, E).unwrapped, E)
    with pytest.raises(DependencyNotFoundError):
        world.get(E)


def test_freeze(provider: ServiceProvider, scope: Scope):
    world.freeze()

    with pytest.raises(FrozenWorldError):
        provider.register(A, scope=scope)


def test_exists(provider: ServiceProvider, scope: Scope):
    provider.register(A, scope=scope)
    assert not provider.exists(object())
    assert provider.exists(A)
    assert provider.exists(Parameterized(A, dict(a=1)))
    assert not provider.exists(Parameterized(B, dict(a=1)))


def test_custom_scope(provider: ServiceProvider):
    dummy_scope = world.scopes.new(name='dummy')

    class MyService:
        pass

    provider.register(MyService, scope=dummy_scope)

    my_service = world.get(MyService)
    assert my_service is world.get(MyService)
    world.scopes.reset(dummy_scope)
    assert my_service is not world.get(MyService)


@pytest.mark.parametrize('klass, scope', [
    pytest.param(object(), None, id='klass'),
    pytest.param(A, object(), id='scope')
])
def test_sanity_checks(provider: ServiceProvider, klass, scope):
    with pytest.raises((AssertionError, TypeError)):
        provider.register(klass, scope=scope)



import inspect
from typing import Callable, Hashable, Optional

# @formatter:off
cimport cython
from cpython.ref cimport PyObject

from antidote.core.container cimport (DependencyResult, FastProvider, header_flag_cacheable, RawContainer)
from .._internal.utils import debug_repr
from ..core import DependencyDebug, Scope
from ..core.exceptions import DependencyNotFoundError

# @formatter:on

cdef extern from "Python.h":
    PyObject *Py_None
    int PyDict_SetItem(PyObject *p, PyObject *key, PyObject *val) except -1
    PyObject *PyDict_GetItem(PyObject *p, PyObject *key)
    PyObject *PyObject_CallObject(PyObject *callable, PyObject *args) except NULL
    int PySet_Contains(PyObject *anyset, PyObject *key) except -1
    void Py_DECREF(PyObject *o)



@cython.final
cdef class IndirectProvider(FastProvider):
    cdef:
        dict __implementations

    def __cinit__(self, dict implementations = None):
        self.__implementations = implementations or dict()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(" \
               f"implementations={list(self.__implementations.keys())})"

    cpdef IndirectProvider clone(self, bint keep_singletons_cache):
        return IndirectProvider.__new__(IndirectProvider, self.__implementations.copy())

    def exists(self, dependency: Hashable) -> bool:
        return (isinstance(dependency, ImplementationDependency)
                and dependency in self.__implementations)

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        cdef:
            ImplementationDependency impl

        if not isinstance(dependency, ImplementationDependency):
            return None

        try:
            target = self.__implementations[dependency]
        except KeyError:
            return None

        impl = <ImplementationDependency> dependency
        if target is None:
            target = impl.implementation()

        return DependencyDebug(debug_repr(impl),
                               scope=Scope.singleton() if impl.permanent else None,
                               wired=[impl.implementation],  # type: ignore
                               dependencies=[target])

    cdef fast_provide(self,
                      PyObject *dependency,
                      PyObject *container,
                      DependencyResult *result):
        cdef:
            PyObject *ptr
            PyObject *target

        ptr = PyDict_GetItem(<PyObject *> self.__implementations, dependency)
        if ptr is NULL:
            return
        elif ptr is not Py_None:
            (<RawContainer> container).fast_get(ptr, result)
            result.header |= header_flag_cacheable()
            if result.value is NULL:
                raise DependencyNotFoundError(<object> ptr)
        else:
            target = PyObject_CallObject(
                <PyObject *> (<ImplementationDependency> dependency).implementation,
                NULL
            )
            (<RawContainer> container).fast_get(target, result)
            if result.value is NULL:
                error = DependencyNotFoundError(<object> target)
                Py_DECREF(target)
                raise error

            if (<ImplementationDependency> dependency).permanent:
                result.header |= header_flag_cacheable()
                PyDict_SetItem(<PyObject *> self.__implementations,
                               dependency,
                               target)
            else:
                result.header = 0

            Py_DECREF(target)

    def register_implementation(self,
                                interface: type,
                                implementation: Callable[[], Hashable],
                                *,
                                permanent: bool
                                ) -> ImplementationDependency:
        assert callable(implementation) \
               and inspect.isclass(interface) \
               and isinstance(permanent, bool)
        impl = ImplementationDependency(interface, implementation, permanent)
        with self._bound_container_ensure_not_frozen():
            self._bound_container_raise_if_exists(impl)
            self.__implementations[impl] = None
            return impl

@cython.final
cdef class ImplementationDependency:
    cdef:
        readonly object interface
        readonly object implementation
        readonly bint permanent
        int _hash

    def __init__(self,
                 interface: Hashable,
                 implementation: Callable[[], Hashable],
                 permanent: bool):
        self.interface = interface
        self.implementation = implementation
        self.permanent = permanent
        self._hash = hash((interface, implementation))

    def __repr__(self) -> str:
        return f"Implementation({self})"

    def __antidote_debug_repr__(self) -> str:
        if self.permanent:
            return f"Permanent implementation: {self}"
        else:
            return f"Implementation: {self}"

    def __str__(self) -> str:
        impl = self.implementation  # type: ignore
        return f"{debug_repr(self.interface)} @ {debug_repr(impl)}"

    # Custom hash & eq necessary to find duplicates
    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: object) -> bool:
        cdef:
            ImplementationDependency imd

        if not isinstance(other, ImplementationDependency):
            return False

        imd = <ImplementationDependency> other
        return (self._hash == imd._hash
                and (self.interface is imd.interface
                     or self.interface == imd.interface)
                and (self.implementation is imd.implementation  # type: ignore
                     or self.implementation == imd.implementation)  # type: ignore
                )  # noqa



import inspect
from typing import Dict, Hashable

# @formatter:off
cimport cython
from cpython.ref cimport PyObject

from antidote.core.container cimport (DependencyResult, FastProvider, header_flag_cacheable, HeaderObject, Scope)
from .._internal.utils import debug_repr
# @formatter:on
from ..core import DependencyDebug

cdef extern from "Python.h":
    PyObject *Py_True
    int PyObject_IsInstance(PyObject *inst, PyObject *cls) except -1
    PyObject *PyDict_GetItem(PyObject *p, PyObject *key)
    PyObject *PyObject_Call(PyObject *callable, PyObject *args, PyObject *kwargs) except NULL
    PyObject *PyObject_CallObject(PyObject *callable, PyObject *args) except NULL

cdef:
    tuple empty_tuple = tuple()


@cython.final
cdef class Parameterized:
    def __init__(self, wrapped: Hashable, parameters: dict):
        assert isinstance(parameters, dict) and len(parameters) > 0
        self.wrapped = wrapped
        self.parameters = parameters

        try:
            self._hash = hash((self.wrapped, tuple(self.parameters.items())))
        except TypeError:
            self._hash = hash((self.wrapped, tuple(self.parameters.keys())))

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return f"Parameterized(dependency={self.wrapped}, parameters={self.parameters})"

    def __antidote_debug_repr__(self):
        return f"{debug_repr(self.wrapped)} with parameters={self.parameters}"

    def __eq__(self, other):
        return (isinstance(other, Parameterized)
                and self._hash == (<Parameterized> other)._hash
                and (self.wrapped is other.wrapped
                     or self.wrapped == other.wrapped)
                and self.parameters == other.parameters)  # noqa

@cython.final
cdef class ServiceProvider(FastProvider):
    """
    Provider managing factories. Also used to register classes directly.
    """
    cdef:
        dict __services

    def __cinit__(self, dict services = None):
        self.__services = services or dict()  # type: Dict[Hashable, HeaderObject]

    def __repr__(self):
        return f"{type(self).__name__}(services={list(self.__services.items())!r})"

    def exists(self, dependency: Hashable) -> bool:
        if isinstance(dependency, Parameterized):
            return dependency.wrapped in self.__services
        return dependency in self.__services

    cpdef ServiceProvider clone(self, bint keep_singletons_cache):
        return ServiceProvider.__new__(ServiceProvider, self.__services.copy())

    def maybe_debug(self, build: Hashable):
        klass = build.wrapped if isinstance(build, Parameterized) else build
        try:
            header = self.__services[klass]
        except KeyError:
            return None
        return DependencyDebug(debug_repr(build),
                               scope=header.to_scope(self._bound_container()),
                               wired=[klass])

    cdef fast_provide(self,
                      PyObject *dependency,
                      PyObject *container,
                      DependencyResult *result):
        cdef:
            PyObject *service
            PyObject *ptr
            tuple args
            object factory
            int scope_or_singleton

        if PyObject_IsInstance(dependency, <PyObject *> Parameterized):
            ptr = PyDict_GetItem(<PyObject *> self.__services,
                                 <PyObject *> (<Parameterized> dependency).wrapped)
            if ptr:
                result.header = (<HeaderObject> ptr).header
                result.value = PyObject_Call(
                    <PyObject *> (<Parameterized> dependency).wrapped,
                    <PyObject *> empty_tuple,
                    <PyObject *> (<Parameterized> dependency).parameters
                )
        else:
            ptr = PyDict_GetItem(<PyObject *> self.__services, dependency)
            if ptr:
                result.header = (<HeaderObject> ptr).header | header_flag_cacheable()
                result.value = PyObject_CallObject(dependency, NULL)

    def register(self, klass: type, *, Scope scope):
        assert inspect.isclass(klass) \
               and (isinstance(scope, Scope) or scope is None)
        with self._bound_container_ensure_not_frozen():
            self._bound_container_raise_if_exists(klass)
            self.__services[klass] = HeaderObject.from_scope(scope)



# pyright: reportUnknownMemberType=false
from __future__ import annotations

from typing import Any, ClassVar, Dict, Optional, Tuple, Type

from .. import API

_ABSTRACT_FLAG = '__antidote_abstract'


@API.private
class FinalMeta(type):
    def __new__(mcs: Type[FinalMeta],
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object]
                ) -> FinalMeta:
        for b in bases:
            if isinstance(b, FinalMeta):
                raise TypeError(f"Type '{b.__name__}' cannot be inherited.")

        return super().__new__(mcs, name, bases, namespace)


@API.private
class AbstractMeta(type):
    def __new__(mcs: Type[AbstractMeta],
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object],
                abstract: bool = False
                ) -> AbstractMeta:
        namespace[_ABSTRACT_FLAG] = abstract
        if not abstract:
            for b in bases:
                if isinstance(b, mcs) and not getattr(b, _ABSTRACT_FLAG):
                    raise TypeError(
                        f"Cannot inherit a service which is not defined abstract. "
                        f"Consider defining {b} abstract by adding 'abstract=True' as a "
                        f"metaclass parameter. If so, Antidote won't use it.")

        return super().__new__(mcs, name, bases, namespace)


@API.private
class Singleton(metaclass=AbstractMeta, abstract=True):
    __instance: ClassVar[Optional[Any]] = None

    def __new__(cls) -> Any:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance



from typing import TYPE_CHECKING

import pytest

from antidote import inject, Provide, world
from antidote.core.exceptions import DoubleInjectionError
from antidote.core.wiring import Wiring, WithWiringMixin

if TYPE_CHECKING:
    pass


class MyService:
    pass


class AnotherService:
    pass


@pytest.fixture(params=['Wiring', 'wire'])
def wire(request):
    kind = request.param
    if kind == 'Wiring':
        def wire(**kwargs):
            return Wiring(**kwargs).wire

        return wire
    else:
        from antidote import wire
        return wire


@pytest.fixture(autouse=True)
def new_world():
    with world.test.empty():
        world.test.singleton({'x': object(),
                              'xx': object(),
                              'y': object(),
                              'z': object(),
                              MyService: MyService(),
                              AnotherService: AnotherService()})
        yield


@pytest.mark.parametrize('kwargs', [
    dict(dependencies=('x', 'y')),
    dict(dependencies=dict(x='x', y='y')),
])
def test_no_strict_validation(wire, kwargs):
    @wire(**kwargs)
    class A:
        def f(self, x):
            return x

        def g(self, x, y):
            return x, y

    a = A()
    assert a.f() == world.get("x")
    assert a.g() == (world.get("x"), world.get("y"))


def test_no_strict_validation_auto_provide(wire):
    @wire(auto_provide=[MyService, AnotherService])
    class A:
        def f(self, x: MyService):
            return x

        def g(self, x: MyService, y: AnotherService):
            return x, y

    a = A()
    assert a.f() == world.get(MyService)
    assert a.g() == (world.get(MyService), world.get(AnotherService))


def test_subclass_classmethod(wire):
    @wire(auto_provide=True)
    class Dummy:
        @classmethod
        def cls_method(cls, x: MyService):
            return cls, x

    assert (Dummy, world.get(MyService)) == Dummy.cls_method()

    class SubDummy(Dummy):
        pass

    assert (SubDummy, world.get(MyService)) == SubDummy.cls_method()


@pytest.mark.parametrize('kwargs, expectation', [
    (dict(methods=object()), pytest.raises(TypeError, match=".*method.*")),
    (dict(methods=[object()]), pytest.raises(TypeError, match="(?i).*method.*")),
    (dict(dependencies=object()), pytest.raises(TypeError, match=".*dependencies.*")),
    (dict(auto_provide=object()), pytest.raises(TypeError, match=".*auto_provide.*")),
    (dict(raise_on_double_injection=object()),
     pytest.raises(TypeError, match=".*raise_on_double_injection.*")),
])
def test_validation(wire, kwargs, expectation):
    with expectation:
        wire(**kwargs)


def test_iterable():
    w = Wiring(methods=iter(['method']),
               auto_provide=iter([MyService]))

    assert isinstance(w.methods, frozenset)
    assert w.methods == {'method'}
    assert isinstance(w.auto_provide, frozenset)
    assert w.auto_provide == {MyService}


def test_default_all_methods(wire):
    @wire()
    class A:
        def __init__(self, x: Provide[MyService]):
            self.x = x

        def __call__(self, x: Provide[MyService]):
            return x

        def method(self, x: Provide[MyService]):
            return x

        @classmethod
        def klass(cls, x: Provide[MyService]):
            return x

        @staticmethod
        def static(x: Provide[MyService]):
            return x

        def _method(self, x: Provide[MyService]):
            return x

        @classmethod
        def _klass(cls, x: Provide[MyService]):
            return x

        @staticmethod
        def _static(x: Provide[MyService]):
            return x

        def __method(self, x: Provide[MyService]):
            return x

        @classmethod
        def __klass(cls, x: Provide[MyService]):
            return x

        @staticmethod
        def __static(x: Provide[MyService]):
            return x

    a = A()
    dummy = world.get(MyService)
    assert a.x is dummy
    assert a() is dummy
    assert a.method() is dummy
    assert a.klass() is dummy
    assert a.static() is dummy
    assert a._method() is dummy
    assert a._klass() is dummy
    assert a._static() is dummy
    assert a._A__method() is dummy
    assert a._A__klass() is dummy
    assert a._A__static() is dummy


def test_methods(wire):
    @wire(methods=['f'], dependencies={'x': MyService})
    class A:
        def f(self, x):
            return x

        def g(self, x):
            return x

    assert A().f() is world.get(MyService)
    with pytest.raises(TypeError):
        A().g()

    with pytest.raises(AttributeError):
        @wire(methods=['f'])
        class A:
            pass


def test_double_injection(wire):
    @wire(methods=['f'])
    class A:
        @inject(auto_provide=True)  # auto_provide to force injection
        def f(self, x: MyService):
            return x

    assert A().f() is world.get(MyService)

    with pytest.raises(DoubleInjectionError):
        @wire(methods=['f'], raise_on_double_injection=True)
        class B:
            @inject(auto_provide=True)  # auto_provide to force injection
            def f(self, x: MyService):
                return x

    @wire()
    class C:
        @inject(auto_provide=True)  # auto_provide to force injection
        def f(self, x: MyService):
            return x

    assert C().f() is world.get(MyService)

    with pytest.raises(DoubleInjectionError):
        @wire(raise_on_double_injection=True)
        class D:
            @inject(auto_provide=True)  # auto_provide to force injection
            def f(self, x: MyService):
                return x


def test_invalid_methods(wire):
    with pytest.raises(TypeError, match='.*not_a_method.*'):
        @wire(methods=['not_a_method'])
        class Dummy:
            not_a_method = 1


@pytest.fixture(params=['method', 'classmethod', 'staticmethod'])
def wired_method_builder(request):
    kind = request.param

    def build(wire, *, annotation=object):
        if kind == 'method':
            @wire
            class A:
                def f(self, x: annotation):
                    return x
        elif kind == 'classmethod':
            @wire
            class A:
                @classmethod
                def f(cls, x: annotation):
                    return x
        else:
            @wire
            class A:
                @staticmethod
                def f(x: annotation):
                    return x
        return A().f

    return build


def test_use_inject_annotation(wire, wired_method_builder):
    f = wired_method_builder(wire())
    with pytest.raises(TypeError):
        f()

    f = wired_method_builder(wire(), annotation=Provide[MyService])
    assert f() is world.get(MyService)


def test_dependencies(wire, wired_method_builder):
    f = wired_method_builder(wire(methods=['f'], dependencies=('y',)))
    assert f() is world.get('y')

    f = wired_method_builder(wire(methods=['f'], dependencies=dict(x='z')))
    assert f() is world.get('z')

    f = wired_method_builder(wire(methods=['f'], dependencies=dict(y='z')))
    with pytest.raises(TypeError):
        f()

    f = wired_method_builder(wire(methods=['f'], dependencies=lambda arg: arg.name * 2))
    assert f() is world.get('xx')


@pytest.mark.parametrize('annotation', [object, MyService])
def test_wiring_auto_provide(wire, wired_method_builder, annotation):
    f = wired_method_builder(wire(methods=['f']),
                             annotation=annotation)
    with pytest.raises(TypeError):
        f()

    # Boolean
    f = wired_method_builder(wire(methods=['f'], auto_provide=False),
                             annotation=annotation)
    with pytest.raises(TypeError):
        f()

    f = wired_method_builder(wire(methods=['f'], auto_provide=True),
                             annotation=annotation)
    if annotation is MyService:
        assert f() is world.get(MyService)
    else:
        with pytest.raises(TypeError):
            f()

    # List
    f = wired_method_builder(wire(methods=['f'], auto_provide=[MyService]),
                             annotation=annotation)
    if annotation is MyService:
        assert f() is world.get(MyService)
    else:
        with pytest.raises(TypeError):
            f()

    class Unknown:
        pass

    f = wired_method_builder(wire(methods=['f'], auto_provide=[Unknown]),
                             annotation=annotation)
    with pytest.raises(TypeError):
        f()

    # Function
    f = wired_method_builder(wire(methods=['f'],
                                  auto_provide=lambda cls: issubclass(cls, MyService)),
                             annotation=annotation)
    if annotation is MyService:
        assert f() is world.get(MyService)
    else:
        with pytest.raises(TypeError):
            f()

    f = wired_method_builder(wire(methods=['f'], auto_provide=lambda cls: False),
                             annotation=annotation)
    with pytest.raises(TypeError):
        f()


def test_complex_wiring(wire):
    @wire(auto_provide=True,
          methods=['g'])
    class A:
        def f(self, x: MyService):
            return x

        def g(self, x: MyService):
            return x

    with pytest.raises(TypeError):
        A().f()

    assert A().g() == world.get(MyService)


def test_class_static_methods(wire):
    @wire(methods=['static', 'klass'],
          auto_provide=True)  # auto_provide to force injection
    class Dummy:
        @staticmethod
        def static(x: MyService):
            pass

        @classmethod
        def klass(cls, x: MyService):
            pass

    assert isinstance(Dummy.__dict__['static'], staticmethod)
    assert isinstance(Dummy.__dict__['klass'], classmethod)


@pytest.mark.parametrize('kwargs', [
    dict(methods={'method', 'test'}),
    dict(dependencies=[MyService]),
    dict(auto_provide=True)
])
def test_copy(kwargs):
    wiring = Wiring(methods=['method'],
                    dependencies=dict(),
                    auto_provide=False)
    copy = wiring.copy(**kwargs)
    for key, value in kwargs.items():
        assert getattr(copy, key) == value


class DummyConf(WithWiringMixin):
    def __init__(self, wiring=None):
        self.wiring = wiring

    def copy(self, wiring: Wiring):
        return DummyConf(wiring)


@pytest.mark.parametrize('kwargs', [
    dict(methods={'method', 'test'}),
    dict(dependencies=[MyService]),
    dict(auto_provide=True),
])
def test_with_wiring(kwargs):
    conf = DummyConf(Wiring(methods=['method'],
                            dependencies=dict(),
                            auto_provide=False))
    copy = conf.with_wiring(**kwargs)
    for key, value in kwargs.items():
        assert getattr(copy.wiring, key) == value

    kwargs.setdefault('methods', {'method'})
    copy = DummyConf().with_wiring(**kwargs)
    for key, value in kwargs.items():
        assert getattr(copy.wiring, key) == value



import pytest

from antidote import Factory, factory, implementation, inject, Provide, service, Service, world
from antidote._implementation import validate_provided_class
from antidote._providers import (FactoryProvider, IndirectProvider,
                                 ServiceProvider)
from antidote.exceptions import DependencyInstantiationError


@pytest.fixture(autouse=True)
def test_world():
    with world.test.empty():
        world.provider(ServiceProvider)
        world.provider(FactoryProvider)
        world.provider(IndirectProvider)
        yield


class Interface:
    pass


def test_default_implementation():
    class A(Interface, Service):
        pass

    class B(Interface, Service):
        pass

    choice = 'a'

    @implementation(Interface)
    def choose():
        return dict(a=A, b=B)[choice]

    assert world.get(Interface @ choose) is world.get(A)
    choice = 'b'
    assert choose() is B
    assert world.get(Interface @ choose) is world.get(A)


@pytest.mark.parametrize('singleton,permanent',
                         [(True, True), (True, False), (False, True), (False, False)])
def test_implementation(singleton: bool, permanent: bool):
    choice = 'a'

    class A(Interface, Service):
        __antidote__ = Service.Conf(singleton=singleton)

    class B(Interface, Service):
        __antidote__ = Service.Conf(singleton=singleton)

    @implementation(Interface, permanent=permanent)
    def choose_service():
        return dict(a=A, b=B)[choice]

    dependency = Interface @ choose_service
    assert isinstance(world.get(dependency), A)
    assert (world.get(dependency) is world.get(A)) is singleton

    choice = 'b'
    assert choose_service() == B
    if permanent:
        assert isinstance(world.get(dependency), A)
        assert (world.get(dependency) is world.get(A)) is singleton
    else:
        assert isinstance(world.get(dependency), B)
        assert (world.get(dependency) is world.get(B)) is singleton


def test_implementation_parameterized_service():
    x = object()

    class A(Interface, Service):
        __antidote__ = Service.Conf(parameters=['test'])

        def __init__(self, **kwargs):
            self.kwargs = kwargs

    @implementation(Interface)
    def impl():
        return A.parameterized(test=x)

    a = world.get(Interface @ impl)
    assert isinstance(a, A)
    assert a.kwargs == dict(test=x)


def test_implementation_with_factory():
    x = object()

    class A(Interface):
        def __init__(self, **kwargs):
            self.kwargs = kwargs

    class BuildA(Factory):
        __antidote__ = Factory.Conf(parameters=['test'])

        def __call__(self, **kwargs) -> A:
            return A(**kwargs)

    @implementation(Interface)
    def impl2():
        return A @ BuildA.parameterized(test=x)

    a = world.get(Interface @ impl2)
    assert isinstance(a, A)
    assert a.kwargs == dict(test=x)


def dummy_choose():
    class A(Interface):
        pass

    return A


@pytest.mark.parametrize('expectation,kwargs,func',
                         [
                             pytest.param(pytest.raises(TypeError, match='.*function.*'),
                                          dict(interface=Interface),
                                          object(),
                                          id='function'),
                             pytest.param(pytest.raises(TypeError, match='.*interface.*'),
                                          dict(interface=object()),
                                          dummy_choose,
                                          id='interface'),
                             pytest.param(pytest.raises(TypeError, match='.*permanent.*'),
                                          {'interface': Interface, 'permanent': object()},
                                          dummy_choose,
                                          id='permanent')
                         ])
def test_invalid_implementation(expectation, kwargs: dict, func):
    with expectation:
        implementation(**kwargs)(func)


def test_invalid_implementation_return_type():
    class B:
        pass

    world.test.singleton(B, 1)

    with world.test.new():
        world.test.singleton(1, 1)

        @implementation(Interface)
        def choose():
            return 1

        world.get(1)
        with pytest.raises(DependencyInstantiationError):
            world.get(Interface @ choose)

    with world.test.new():
        world.test.singleton(B, 1)

        @implementation(Interface)
        def choose2():
            return B

        world.get(B)
        with pytest.raises(DependencyInstantiationError):
            world.get(Interface @ choose2)

    with world.test.new():
        class C(Service):
            __antidote__ = Service.Conf(parameters=['test'])

            def __init__(self, **kwargs):
                self.kwargs = kwargs

        @implementation(Interface)
        def impl():
            return C.parameterized(test=1)

        world.get(C.parameterized(test=1))
        with pytest.raises(DependencyInstantiationError):
            world.get(Interface @ impl)

    with world.test.new():
        class D:
            def __init__(self, **kwargs):
                self.kwargs = kwargs

        class BuildD(Factory):
            __antidote__ = Factory.Conf(parameters=['test'])

            def __call__(self, **kwargs) -> D:
                return D(**kwargs)

        @implementation(Interface)
        def impl2():
            return D @ BuildD.parameterized(test=1)

        world.get(D @ BuildD.parameterized(test=1))
        with pytest.raises(DependencyInstantiationError):
            world.get(Interface @ impl2)


def test_invalid_implementation_dependency():
    class Interface:
        pass

    class A(Interface, Service):
        pass

    @implementation(Interface)
    def current_interface():
        return A

    with pytest.raises(ValueError, match=".*interface.*"):
        A @ current_interface


def test_getattr():
    class Interface:
        pass

    class A(Interface, Service):
        pass

    def current_interface():
        return A

    current_interface.hello = 'world'

    build = implementation(Interface)(current_interface)
    assert build.hello == 'world'

    build.new_hello = 'new_world'
    assert build.new_hello == 'new_world'


def test_validate_provided_class():
    class Interface:
        pass

    with pytest.raises(TypeError):
        validate_provided_class(object(), expected=Interface)

    class A(Interface, Service):
        __antidote__ = Service.Conf(parameters=['a'])

    class B(Service):
        __antidote__ = Service.Conf(parameters=['a'])

    validate_provided_class(A, expected=Interface)
    validate_provided_class(B, expected=B)
    with pytest.raises(TypeError):
        validate_provided_class(B, expected=Interface)

    validate_provided_class(A.parameterized(a=1), expected=Interface)
    validate_provided_class(B.parameterized(a=1), expected=B)
    with pytest.raises(TypeError):
        validate_provided_class(B.parameterized(a=1), expected=Interface)

    @implementation(Interface)
    def choose_a():
        return A

    @implementation(B)
    def choose_b():
        return B

    validate_provided_class(Interface @ choose_a, expected=Interface)
    validate_provided_class(B @ choose_b, expected=B)
    with pytest.raises(TypeError):
        validate_provided_class(B @ choose_b, expected=Interface)


def test_validate_provided_class_factory():
    class Interface:
        pass

    class A(Interface):
        pass

    class B:
        pass

    with world.test.new():
        @factory
        def build_a() -> A:
            return A()

        @factory
        def build_b() -> B:
            return B()

        validate_provided_class(A @ build_a, expected=Interface)
        validate_provided_class(B @ build_b, expected=B)
        with pytest.raises(TypeError):
            validate_provided_class(B @ build_b, expected=Interface)

    with world.test.new():
        class BuildA(Factory):
            __antidote__ = Factory.Conf(parameters=['a'])

            def __call__(self, **kwargs) -> A:
                return A(**kwargs)

        class BuildB(Factory):
            __antidote__ = Factory.Conf(parameters=['a'])

            def __call__(self, **kwargs) -> B:
                return B(**kwargs)

        validate_provided_class(A @ BuildA, expected=Interface)
        validate_provided_class(B @ BuildB, expected=B)
        with pytest.raises(TypeError):
            validate_provided_class(B @ BuildB, expected=Interface)

        validate_provided_class(A @ BuildA.parameterized(a=1), expected=Interface)
        validate_provided_class(B @ BuildB.parameterized(a=1), expected=B)
        with pytest.raises(TypeError):
            validate_provided_class(B @ BuildB.parameterized(a=1), expected=Interface)


def test_default_injection():
    class MyService(Service):
        pass

    class Interface:
        pass

    class A(Interface, Service):
        pass

    injected = None

    @implementation(Interface)
    def choose_a(s: Provide[MyService]):
        nonlocal injected
        injected = s
        return A

    assert world.get(Interface @ choose_a) is world.get(A)
    assert injected is world.get(MyService)


def test_double_injection():
    class B:
        pass

    world.test.singleton(B, object())

    class Interface:
        pass

    class A(Interface, Service):
        pass

    injected = None

    @implementation(Interface)
    @inject(auto_provide=True)
    def choose_a(s: B):
        nonlocal injected
        injected = s
        return A

    assert world.get(Interface @ choose_a) is world.get(A)
    assert injected is world.get(B)


def test_source_notation():
    @service
    class A(Interface):
        pass

    @implementation(Interface)
    def choose():
        return A

    assert world.get(Interface, source=choose) is world.get(A)

    @inject
    def f(a: Interface = inject.me(source=choose)) -> Interface:
        return a

    assert f() is world.get(A)

    with pytest.raises(ValueError, match="Unsupported interface.*"):
        world.get(A, source=choose)



import enum
import typing

from typing_extensions import Protocol

from .debug import debug_repr, short_id
from .immutable import FinalImmutable, Immutable
from .meta import AbstractMeta, FinalMeta
from .. import API

Im = typing.TypeVar('Im', bound=Immutable)


@API.private
class Default(enum.Enum):
    sentinel = enum.auto()


@API.private
class Copy(enum.Enum):
    IDENTICAL = enum.auto()

    @staticmethod
    def immutable(current: Im, **kwargs: object) -> Im:
        return type(current)(**{
            attr: getattr(current, attr) if value is Copy.IDENTICAL else value
            for attr, value in kwargs.items()
        })


__all__ = ['debug_repr', 'short_id', 'FinalImmutable', 'Immutable', 'AbstractMeta',
           'FinalMeta', 'API', 'Default', 'Copy']

# 3.8+
if hasattr(typing, 'runtime_checkable'):

    @API.private
    def enforce_type_if_possible(obj: object, tpe: object) -> None:
        if not isinstance(tpe, type):
            return

        # inspired by how `typing.runtime_checkable` checks for a protocol
        if issubclass(tpe, typing.Generic) and getattr(tpe, "_is_protocol", False):  # type: ignore
            if getattr(tpe, "_is_runtime_protocol", False) and not isinstance(obj, tpe):
                raise TypeError(f"Dependency value {obj} is not an instance of {tpe}, "
                                f"but a {type(obj)}")
        elif not isinstance(obj, tpe):
            raise TypeError(f"Dependency value {obj} is not an instance of {tpe}, "
                            f"but a {type(obj)}")
else:
    ProtocolMeta = type(Protocol)

    @API.private
    def enforce_type_if_possible(obj: object, tpe: object) -> None:
        if not isinstance(tpe, type):
            return

        # inspired by how `typing_extensions.runtime_checkable` checks for a protocol
        if isinstance(tpe, ProtocolMeta) and getattr(tpe, "_is_protocol", False):
            if getattr(tpe, "_is_runtime_protocol", False) and not isinstance(obj, tpe):
                raise TypeError(f"Dependency value {obj} is not an instance of {tpe}, "
                                f"but a {type(obj)}")
        elif not isinstance(obj, tpe):
            raise TypeError(f"Value {obj!r} is not an instance of {tpe!r}, "
                            f"but a {type(obj)!r}")



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Callable, cast, Generic, Hashable, Iterator, Optional, TypeVar  # noqa: F401

from typing_extensions import final, TypeAlias

from ._provider import FREEZE_ATTR_NAME, ProviderMeta
from .container import Container, DependencyValue, RawProvider
from .exceptions import DebugNotAvailableError
from .utils import DependencyDebug
from .._internal import API
from .._internal.utils import debug_repr

T = TypeVar('T', bound=Hashable)
M = TypeVar('M', bound=Callable[..., object])
AnyProvider: TypeAlias = 'Provider[Any]'
P = TypeVar('P', bound=AnyProvider)


@API.public
def does_not_freeze(method: M) -> M:
    """
    Decorated method won't freeze when :py:func:`~antidote.world.freeze` is called. All
    others will, with the exceptions of base methods defined in :py:class`.Provider`.

    Beware that is must be the last decorator if you use multiple ones.
    """
    setattr(method, FREEZE_ATTR_NAME, False)
    return method


@API.public
class Provider(RawProvider, Generic[T],
               metaclass=ProviderMeta,
               # reserved
               abstract=True):
    """
    Abstract Base class for a provider.

    Consider using :py:class:`~.StatelessProvider` if you do not have any state
    to keep.

    .. doctest:: core_Provider

        >>> from typing import Hashable, Optional
        >>> from antidote import world, Scope
        >>> from antidote.core import Provider, Container, \\
        ...     DependencyValue, does_not_freeze
        >>> @world.provider
        ... class SquareProvider(Provider[int]):
        ...     def __init__(self, registered: set = None):
        ...         super().__init__()
        ...         self._registered = registered or set()
        ...
        ...     def register_numbers(self, number: int):
        ...         self._assert_not_duplicate(number)
        ...         self._registered |= {number}
        ...
        ...     def exists(self, dependency: Hashable):
        ...         return isinstance(dependency, int) and dependency in self._registered
        ...
        ...     def provide(self, dependency: int, container: Container
        ...                 ) -> DependencyValue:
        ...         return DependencyValue(self._square(dependency),
        ...                                   scope=Scope.singleton())
        ...
        ...     # we don't want this method to fail when world freezes
        ...     # and it does not modify any state, so we're safe !
        ...     @does_not_freeze
        ...     def _square(self, n: int) -> int:
        ...         return n ** 2
        ...
        ...     def clone(self, keep_singletons_cache: bool) -> 'SquareProvider':
        ...         return SquareProvider(self._registered.copy())
        ...
        >>> # To modify your provider, retrieve it from world
        ... world.get[SquareProvider]().register_numbers(9)
        >>> world.get[SquareProvider]().register_numbers(7)
        >>> world.get[int](9)
        81
        >>> # If you freeze world, you cannot register any dependencies
        ... world.freeze()
        >>> from antidote.exceptions import FrozenWorldError
        >>> try:
        ...     world.get[SquareProvider]().register_numbers(10)
        ... except FrozenWorldError:
        ...     print("Frozen world !")
        Frozen world !
        >>> # But you can still retrieve previously defined ones
        ... world.get[int](9)
        81
        >>> # even if never called before
        ... world.get[int](7)
        49


    .. warning::
        This is most advanced feature of Antidote and allows you to extend Antidote's
        behavior. So be careful with it, it will impact the whole library. There are
        several rules to respect:

        1. Different providers MUST provide strictly different dependencies.
        2. You MUST NOT use :py:mod:`~antidote.world`. If you need a dependency, rely on
           the provided :py:class:`~.core.container.Container`.
        3. Methods will automatically freeze except those marked with the decorator
           :py:func:`~.does_not_freeze`. Methods changing dependencies *definitions* MUST
           NOT use :py:func:`~.does_not_freeze`, others may.
        4. You may cache singletons by yourself, but you need to clean your cache when
           :py:meth:`~.clone` is called with :code:`keep_singletons_cache=False`.
    """
    __antidote__ = None  # reserved

    def clone(self: P, keep_singletons_cache: bool) -> P:
        """
        If you have no internal state, consider implementing
        :py:class:`~.StatelessProvider` instead.

        Used by the different test utilities.

        Args:
            keep_singletons_cache: Whether cached singletons, if any, should be kept or
                discarded.

        Returns:
            A deep copy of the current provider. It should always be a new instance.
        """
        raise NotImplementedError()

    def exists(self, dependency: Hashable) -> bool:
        """
        Check whether dependency exists in the current provider. It is recommended to be
        relatively fast as this function will often be called for all _providers. Among
        others it is used to check for duplicates and ensure that it is providable.

        Args:
            dependency: Dependency to check.

        Returns:
            bool: Whether dependency has been registered in this provider or not.
        """
        raise NotImplementedError()

    def provide(self, dependency: T, container: Container) -> DependencyValue:
        """
        Method called by the :py:class:`~.core.container.Container` when
        searching for a dependency. Be sure that the dependency space of your
        _providers don't intersect ! This function will only be called if
        :py:meth:`.exists` succeeded on the :code:`dependency`.

        If you need to access other dependencies, you MUST use the provided
        :code:`container` argument and NEVER :py:mod:`~antidote.world`.

        It is always called within a thread-safe (locked) environment, so you
        don't need to worry about it. You also don't need to worry about
        handling singletons, the :py:class:`~.core.container.Container` will
        handle it for you.

        Be careful to be consistent with :py:meth:`.exists`.

        Args:
            dependency: The dependency to be provided by the provider. It will have passed
                :py:meth:`.exists`.
            container: current container which may use to retrieve other dependencies.

        Returns:
            The requested instance wrapped in a
            :py:class:`~.core.container.DependencyValue`. If the dependency is a
            singleton, you MUST specify it with :code:`singleton=True`.
        """
        raise NotImplementedError("Either implement provide()"
                                  "or override maybe_provide()")

    def debug(self, dependency: T) -> DependencyDebug:
        """
        Optional support for :py:mod:`.world.debug`. If not implemented, debug information
        will not be provided for dependencies.

        Args:
            dependency: The dependency for which debug information should be provided. It
                will have passed :py:meth:`.exists`.

        Returns:
            A short information text on the dependency, whether it's a singleton and
            everything that has been wired for the dependency and its respective
            dependencies if any.

        """
        raise DebugNotAvailableError("Either implement debug() or override maybe_debug()")

    def maybe_provide(self,
                      dependency: Hashable,
                      container: Container
                      ) -> Optional[DependencyValue]:
        """
        **Expert feature**

        :py:meth:`.maybe_provide` MUST be consistent with :py:meth:`.exists`. It should
        return

        Args:
            dependency: The dependency to be provided by the provider.
            container: current container which may use to retrieve other dependencies.

        Returns:
            The requested instance wrapped in a
            :py:class:`~.core.container.DependencyValue` if available or
            :py:obj:`None`. If the dependency is a singleton, you MUST specify it
            with :code:`singleton=True`.

        """
        if self.exists(dependency):
            return self.provide(cast(T, dependency), container)
        return None

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        """
        **Expert feature**

        :py:meth:`.maybe_debug` MUST be consistent with :py:meth:`.exists`.

        Args:
            dependency: The dependency for which debug information should be provided. It
                will have passed :py:meth:`.exists`.

        Returns:
            A short information text on the dependency, whether it's a singleton and
            everything that has been wired for the dependency and its respective
            dependencies if any.

        """
        if self.exists(dependency):
            try:
                return self.debug(cast(T, dependency))
            except DebugNotAvailableError:
                import warnings
                warnings.warn(f"Debug information for {debug_repr(dependency)} "
                              f"not available in {type(self)}")
        return None

    @does_not_freeze
    @final
    def _assert_not_duplicate(self, dependency: object) -> None:
        """
        To be used whenever registering new dependencies to check that a dependency has
        not been declared before.

        Args:
            dependency: Dependency to check

        Raises:
            DuplicateDependencyError
        """
        self._bound_container_raise_if_exists(dependency)

    @does_not_freeze
    @final
    @contextmanager
    def _container_lock(self) -> Iterator[None]:
        """
        Context manager ensuring thread-safety regarding dependency instantiation and
        declaration over all providers. Typically useful if you want to change some
        mutable state in your provider.

        """
        with self._bound_container_locked():
            yield


@API.public
class StatelessProvider(Provider[T], abstract=True):
    """
    Abstract stateless :py:class:`~.Provider` which
    implements :py:meth:`~.clone`.

    .. doctest:: core_StatelessProvider

        >>> from typing import Hashable, Optional
        >>> from antidote import world, Scope
        >>> from antidote.core import StatelessProvider, Container, DependencyValue
        >>> @world.provider
        ... class SquareProvider(StatelessProvider[int]):
        ...     def exists(self, dependency: Hashable) -> bool:
        ...         return isinstance(dependency, int)
        ...
        ...     def provide(self, dependency: int, container: Container
        ...                 ) -> Optional[DependencyValue]:
        ...         return DependencyValue(dependency ** 2, scope=Scope.singleton())
        >>> world.get(9)
        81
    """

    @final
    def clone(self, keep_singletons_cache: bool) -> StatelessProvider[T]:
        return type(self)()



import threading
from collections import deque
from contextlib import contextmanager
from typing import (Any, Callable, Deque, Dict, Hashable, List, Mapping, Optional,
                    Tuple, Type)
from weakref import ref

# @formatter:off
cimport cython
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from cpython.ref cimport PyObject
from fastrlock.rlock cimport create_fastrlock, lock_fastrlock, unlock_fastrlock

from antidote._internal.stack cimport DependencyStack
from .exceptions import (DependencyCycleError, DependencyInstantiationError,
                         DependencyNotFoundError, DuplicateDependencyError,
                         FrozenWorldError)
# @formatter:on

cdef extern from "Python.h":
    PyObject *PyList_GET_ITEM(PyObject *list, Py_ssize_t i)
    Py_ssize_t PyList_Size(PyObject *list)
    Py_ssize_t PyTuple_GET_SIZE(PyObject *p)
    int PyDict_SetItem(PyObject *p, PyObject *key, PyObject *val) except -1
    PyObject *PyDict_GetItem(PyObject *p, PyObject *key)
    void Py_INCREF(PyObject *o)
    void Py_DECREF(PyObject *o)


##############
# Dependency #
##############

DEF HEADER_FLAG_NO_SCOPE = 0
DEF HEADER_FLAG_SINGLETON = 1
DEF HEADER_FLAG_HAS_SCOPE = 2
DEF HEADER_FLAG_CACHEABLE = 4

cdef inline ScopeId header_get_scope_id(Header header):
    return header >> 8

cdef inline Header header_scope(ScopeId scope_id):
    return HEADER_FLAG_HAS_SCOPE | ((scope_id & 0xFF) << 8)

cdef bint header_is_singleton(Header header):
    return header & HEADER_FLAG_SINGLETON

cdef bint header_has_scope(Header header):
    return header & HEADER_FLAG_HAS_SCOPE

cdef bint header_is_cacheable(Header header):
    return header & HEADER_FLAG_CACHEABLE

cdef Header header_flag_singleton():
    return HEADER_FLAG_SINGLETON

cdef Header header_flag_no_scope():
    return HEADER_FLAG_NO_SCOPE

cdef Header header_flag_cacheable():
    return HEADER_FLAG_CACHEABLE

@cython.final
cdef class Scope:
    def __init__(self, str name):
        self.name = name

    def __repr__(self):
        return f"Scope(name='{self.name}')"

    @staticmethod
    def singleton():
        return _SCOPE_SINGLETON

    @staticmethod
    def sentinel():
        return _SCOPE_SENTINEL

_SCOPE_SINGLETON = Scope('singleton')
_SCOPE_SENTINEL = Scope('__sentinel__')

cdef class HeaderObject:
    """
    Utility wrapper for maybe_debug() and in case the header needs to be wrapped inside
    an object (dict values in service for example)
    """
    def __cinit__(self, Header header):
        self.header = header

    @staticmethod
    cdef HeaderObject from_scope(Scope scope):
        if scope is None:
            return HeaderObject.__new__(HeaderObject, HEADER_FLAG_NO_SCOPE)
        if scope is _SCOPE_SINGLETON:
            return HeaderObject.__new__(HeaderObject, HEADER_FLAG_SINGLETON)
        return HeaderObject.__new__(HeaderObject, header_scope(scope.id))

    def to_scope(self, RawContainer container) -> Optional[Scope]:
        scope = None
        if self.is_singleton():
            scope = _SCOPE_SINGLETON
        elif self.has_scope():
            scope = container.get_scope(self.get_scope_id())
        return scope

    def is_singleton(self):
        return header_is_singleton(self.header)

    def has_scope(self):
        return header_has_scope(self.header)

    def get_scope_id(self):
        return header_get_scope_id(self.header)

    def is_cacheable(self):
        return header_is_cacheable(self.header)

@cython.freelist(64)
@cython.final
cdef class DependencyValue:
    def __cinit__(self, value, *, Scope scope = None):
        assert scope is not _SCOPE_SENTINEL
        self.unwrapped = value
        self.scope = scope

    def __repr__(self):
        return f"DependencyValue(unwrapped={self.unwrapped}, scope={self.scope})"

    def __eq__(self, other):
        return isinstance(other, DependencyValue) \
               and self.unwrapped == other.unwrapped \
               and self.scope == other.scope

    def is_singleton(self) -> bool:
        return self.scope is _SCOPE_SINGLETON

    cdef to_result(self, DependencyResult *result):
        result.header = HeaderObject.from_scope(self.scope).header
        result.value = <PyObject *> self.unwrapped
        Py_INCREF(result.value)

    @staticmethod
    cdef DependencyValue from_result(RawContainer container, DependencyResult *result):
        scope = HeaderObject(result.header).to_scope(container)
        value = <object> result.value
        print(scope, value)
        Py_DECREF(result.value)
        return DependencyValue.__new__(DependencyValue,
                                       value,
                                       scope=scope)

_DEFAULT_SENTINEL = object()

############
# PROVIDER #
############

cdef class Container:
    def get(self, dependency: Hashable, default: object = _DEFAULT_SENTINEL):
        raise NotImplementedError()  # pragma: no cover

    def provide(self, dependency: Hashable):
        raise NotImplementedError()  # pragma: no cover

cdef class RawProvider:
    """
    Contrary to the Python implementation, the container does not rely on provide() but
    on fast_provide(). This is done to improve performance by avoiding object creation.
    """

    def __cinit__(self):
        self._container_ref = None

    # Hack to support custom cloned type for FactoryProvider.
    cdef object provider_type(self):
        return type(self)

    cpdef RawProvider clone(self, bint keep_singletons_cache):
        raise NotImplementedError()  # pragma: no cover

    def exists(self, dependency: Hashable) -> bool:
        raise NotImplementedError()  # pragma: no cover

    def maybe_debug(self, dependency: Hashable):
        raise NotImplementedError()  # pragma: no cover

    def maybe_provide(self,
                      dependency: Hashable,
                      container: Container) -> Optional[DependencyValue]:
        raise NotImplementedError()

    cdef fast_provide(self,
                      PyObject *dependency,
                      PyObject *container,
                      DependencyResult *result):
        cdef:
            DependencyValue dependency_instance
        dependency_instance = self.maybe_provide(<object> dependency,
                                                 <Container> container)
        if dependency_instance is not None:
            dependency_instance.to_result(result)

    @contextmanager
    def _bound_container_ensure_not_frozen(self):
        container = self._bound_container()
        if container is not None:
            with container.ensure_not_frozen():
                yield
        else:
            yield

    @contextmanager
    def _bound_container_locked(self, *, freezing: bool = False):
        container = self._bound_container()
        if container is not None:
            with container.locked(freezing=freezing):
                yield
        else:
            yield

    def _bound_container_raise_if_exists(self, dependency: Hashable):
        container = self._bound_container()
        if container is not None:
            container.raise_if_exists(dependency)
        else:
            if self.exists(dependency):
                raise DuplicateDependencyError(
                    f"{dependency} has already been registered in {type(self)}")

    cdef RawContainer _bound_container(self):
        if self._container_ref is not None:
            container = self._container_ref()
            assert container is not None, "Associated container does not exist anymore."
            return container
        return None

    @property
    def is_registered(self):
        return self._container_ref is not None

cdef class FastProvider(RawProvider):
    def maybe_provide(self,
                      dependency: Hashable,
                      container: Container) -> Optional[DependencyValue]:
        cdef:
            DependencyResult result
        result.value = NULL
        self.fast_provide(<PyObject *> dependency, <PyObject *> container, &result)
        if result.value:
            return DependencyValue.from_result(self._bound_container(), &result)
        return None

    cdef fast_provide(self,
                      PyObject *dependency,
                      PyObject *container,
                      DependencyResult *result):
        raise NotImplementedError()

#############
# CONTAINER #
#############

cdef class RawContainer(Container):
    """
    Behaves the same as the Python implementation but with additional optimizations:

    - singletons clock: Avoid testing twice the singleton dictionary it hasn't changed
                        since the first check.
    - cache: Singletons are caching in simplified and faster hash map. For other
             dependencies, if the header specify cacheable, the provider and scope will
             be cached.
    """

    def __init__(self):
        self._dependency_stack = DependencyStack()
        self._instantiation_lock = create_fastrlock()
        self._registration_lock = threading.RLock()
        self.__frozen = False
        self.__providers = list()  # type: List[RawProvider]
        self.__singletons = dict()  # type: dict
        self.__scopes = []
        self.__scope_dependencies = []  # type: List[dict]

        # Cython optimizations
        self.__singletons_clock = 0
        self.__cache = DependencyCache()

    def __repr__(self):
        return f"{type(self).__name__}(__providers={', '.join(map(str, self.__providers))})"

    @staticmethod
    def with_same_providers_and_scopes(RawContainer original):
        cdef:
            RawProvider provider

        container = RawContainer()
        for provider in original.providers:
            container.add_provider(provider.provider_type())
        container.__scopes = original.__scopes.copy()
        container.__scope_dependencies = [dict() for _ in range(len(original.__scopes))]
        return container

    @property
    def scopes(self):
        return self.__scopes.copy()

    @property
    def providers(self) -> List[RawProvider]:
        return self.__providers.copy()

    @contextmanager
    def locked(self, *, freezing: bool = False):
        assert isinstance(freezing, bool)
        with self._registration_lock, self._instantiation_lock:
            if freezing and self.__frozen:
                raise FrozenWorldError()
            yield

    @contextmanager
    def ensure_not_frozen(self):
        with self._registration_lock:
            if self.__frozen:
                raise FrozenWorldError()
            yield

    def freeze(self):
        with self._registration_lock:
            if self.__frozen:
                raise FrozenWorldError("Container is already frozen !")
            self.__frozen = True

    def add_provider(self, provider_cls: Type[RawProvider]):
        cdef:
            RawProvider provider, p

        with self.locked(freezing=True):
            for p in self.__providers:
                assert provider_cls != p.provider_type()
            provider = provider_cls()
            provider._container_ref = ref(self)
            self.__providers.append(provider)
            (<DependencyCache> self.__cache).set(<PyObject *> provider_cls,
                                                 HEADER_FLAG_SINGLETON,
                                                 <PyObject *> provider)
            self.__singletons_clock += 1

    def add_singletons(self, dependencies: Mapping):
        with self.locked(freezing=True):
            for k, v in dependencies.items():
                self.raise_if_exists(k)
            for k, v in dependencies.items():
                (<DependencyCache> self.__cache).set(<PyObject *> k,
                                                     HEADER_FLAG_SINGLETON,
                                                     <PyObject *> v)
            self.__singletons.update(dependencies)
            self.__singletons_clock += 1

    def create_scope(self, str name):
        cdef:
            Scope s = Scope(name)
        with self.locked(freezing=True):
            s.id = <ScopeId> (1 + len(self.__scopes))
            assert s.id <= 0xFF
            assert all(s.name != name for s in self.__scopes)
            self.__scopes.append(s)
            self.__scope_dependencies.append(dict())
        return s

    def reset_scope(self, Scope scope):
        with self._instantiation_lock:
            self.__scope_dependencies[scope.id - 1] = dict()

    cdef Scope get_scope(self, ScopeId scope_id):
        return <Scope> self.__scopes[scope_id - 1]

    def raise_if_exists(self, dependency: Hashable):
        cdef:
            RawProvider provider

        with self._registration_lock:
            if dependency in self.__singletons:
                raise DuplicateDependencyError(
                    f"{dependency!r} has already been defined as a singleton pointing "
                    f"to {self.__singletons[dependency]}")

            for provider in self.__providers:
                if provider.exists(dependency):
                    debug = provider.maybe_debug(dependency)
                    message = f"{dependency!r} has already been declared " \
                              f"in {provider.provider_type()}"
                    if debug is None:
                        raise DuplicateDependencyError(message)
                    else:
                        raise DuplicateDependencyError(message + f"\n{debug.info}")

    def clone(self,
              *,
              keep_singletons: bool = False,
              keep_scopes: bool = False) -> RawContainer:
        cdef:
            RawContainer clone
            RawProvider p, p_clone

        with self.locked():
            clone = OverridableRawContainer()
            clone.__frozen = True
            if keep_singletons:
                clone.__singletons = self.__singletons.copy()

            clone.__scopes = self.__scopes
            clone.__scope_dependencies = [
                d.copy() if keep_scopes else dict()
                for d in self.__scope_dependencies
            ]

            for p in self.__providers:
                p_clone = p.clone(keep_singletons)
                if p_clone is p or p_clone._container_ref is not None:
                    raise RuntimeError("A Provider should always return a fresh "
                                       "instance when copy() is called.")
                p_clone._container_ref = ref(clone)
                clone.__providers.append(p_clone)
                clone.__singletons[p.provider_type()] = p_clone

            return clone

    def debug(self, dependency: Hashable):
        from .._internal.utils.debug import debug_repr
        from .utils import DependencyDebug

        with self._registration_lock:
            for p in self.__providers:
                debug = p.maybe_debug(dependency)
                if debug is not None:
                    return debug
            try:
                value = self.__singletons[dependency]
                return DependencyDebug(f"Singleton: {debug_repr(dependency)} "
                                       f"-> {value!r}",
                                       scope=Scope.singleton())
            except KeyError:
                raise DependencyNotFoundError(dependency)

    def provide(self, dependency: Hashable):
        cdef:
            DependencyResult result

        self.fast_get(<PyObject *> dependency, &result)
        if result.value:
            return DependencyValue.from_result(self, &result)
        raise DependencyNotFoundError(dependency)

    def get(self, dependency: Hashable, default: object = _DEFAULT_SENTINEL):
        cdef:
            DependencyResult result
            object obj

        self.fast_get(<PyObject *> dependency, &result)
        if result.value:
            obj = <object> result.value
            Py_DECREF(result.value)
            return obj
        if default is not _DEFAULT_SENTINEL:
            return default
        raise DependencyNotFoundError(dependency)

    # No ownership from here on. You MUST keep a valid reference to dependency.
    # result.value will be initialized to NULL here, so it doesn't need to be done
    # anywhere else as everything needs to go through fast_get.
    cdef fast_get(self, PyObject *dependency, DependencyResult *result):
        cdef:
            CacheValue *value
            unsigned long clock
            PyObject *ptr

        result.value = NULL
        value = (<DependencyCache> self.__cache).get(dependency)
        if value:
            if value.header & HEADER_FLAG_SINGLETON:
                result.header = value.header
                result.value = value.ptr
                Py_INCREF(result.value)
            else:
                self.__safe_cache_provide(dependency, result, value)
        else:
            clock = self.__singletons_clock
            ptr = PyDict_GetItem(<PyObject *> self.__singletons, dependency)
            if ptr:
                result.header = HEADER_FLAG_SINGLETON
                result.value = ptr
                Py_INCREF(result.value)
            else:
                self.__safe_provide(dependency, result, clock)

    cdef __safe_cache_provide(self,
                              PyObject *dependency,
                              DependencyResult *result,
                              CacheValue *cached):
        cdef:
            ScopeId scope_id
            PyObject *value
            object lock = self._instantiation_lock
            PyObject *stack = <PyObject *> self._dependency_stack
            PyObject *scope_dependencies = <PyObject *> self.__scope_dependencies

        lock_fastrlock(lock, -1, True)

        if cached.header & HEADER_FLAG_HAS_SCOPE:
            scope_id = header_get_scope_id(cached.header)
            value = PyDict_GetItem(
                PyList_GET_ITEM(scope_dependencies, scope_id - 1),
                dependency
            )
            if value:
                unlock_fastrlock(lock)
                result.header = cached.header
                result.value = value
                Py_INCREF(result.value)
                return

        if 0 != (<DependencyStack> stack).push(dependency):
            error = (<DependencyStack> stack).reset_with_error(dependency)
            unlock_fastrlock(lock)
            raise error

        try:
            (<RawProvider> cached.ptr).fast_provide(dependency, <PyObject *> self, result)
            assert result.value, "Once cached, a dependency must always be providable"
            cached.header = result.header
            if result.header & HEADER_FLAG_SINGLETON:
                PyDict_SetItem(<PyObject *> self.__singletons, dependency, result.value)
                self.__singletons_clock += 1
                Py_DECREF(cached.ptr)
                cached.ptr = result.value
                Py_INCREF(result.value)
            elif result.header & HEADER_FLAG_HAS_SCOPE:
                scope_id = header_get_scope_id(result.header)
                PyDict_SetItem(
                    PyList_GET_ITEM(scope_dependencies, scope_id - 1),
                    dependency,
                    result.value
                )
        except Exception as error:
            new_error = handle_error(dependency, stack, error)
            if new_error is not error:
                raise new_error from error
            else:
                raise
        finally:
            (<DependencyStack> stack).pop()
            unlock_fastrlock(lock)

    cdef __safe_provide(self,
                        PyObject *dependency,
                        DependencyResult *result,
                        unsigned long singletons_clock):
        cdef:
            PyObject *value
            PyObject *provider
            ScopeId scope_id
            Exception error
            object lock = self._instantiation_lock
            PyObject *singletons = <PyObject *> self.__singletons
            PyObject *stack = <PyObject *> self._dependency_stack
            PyObject *providers
            PyObject *scope_dependencies
            size_t i

        lock_fastrlock(lock, -1, True)

        # If anything changed in the singletons the clock would be different
        # otherwise no need to re-check the dictionary.
        if singletons_clock < self.__singletons_clock:
            value = PyDict_GetItem(singletons, dependency)
            if value:
                unlock_fastrlock(lock)
                result.header = HEADER_FLAG_SINGLETON
                result.value = value
                Py_INCREF(result.value)
                return

        scope_dependencies = <PyObject *> self.__scope_dependencies
        for i in range(<size_t> PyList_Size(scope_dependencies)):
            value = PyDict_GetItem(PyList_GET_ITEM(scope_dependencies, i), dependency)
            if value:
                unlock_fastrlock(lock)
                result.header = header_scope(i + 1)
                result.value = value
                Py_INCREF(result.value)
                return

        if 0 != (<DependencyStack> stack).push(dependency):
            error = (<DependencyStack> stack).reset_with_error(dependency)
            unlock_fastrlock(lock)
            raise error

        try:
            providers = <PyObject *> self.__providers
            for i in range(<size_t> PyList_Size(providers)):
                provider = PyList_GET_ITEM(providers, i)
                (<RawProvider> provider).fast_provide(
                    dependency,
                    <PyObject *> self,
                    result
                )
                if result.value:
                    if result.header & HEADER_FLAG_SINGLETON:
                        PyDict_SetItem(singletons, dependency, result.value)
                        self.__singletons_clock += 1
                        (<DependencyCache> self.__cache).set(dependency,
                                                             result.header,
                                                             result.value)
                    else:
                        if result.header & HEADER_FLAG_CACHEABLE:
                            (<DependencyCache> self.__cache).set(dependency,
                                                                 result.header,
                                                                 provider)
                        if result.header & HEADER_FLAG_HAS_SCOPE:
                            scope_id = header_get_scope_id(result.header)
                            PyDict_SetItem(
                                PyList_GET_ITEM(scope_dependencies, scope_id - 1),
                                dependency,
                                result.value
                            )
                    return

        except Exception as error:
            new_error = handle_error(dependency, stack, error)
            if new_error is not error:
                raise new_error from error
            else:
                raise
        finally:
            (<DependencyStack> stack).pop()
            unlock_fastrlock(lock)

cdef inline object handle_error(PyObject *dependency, PyObject *stack, object error):
    if isinstance(error, DependencyCycleError):
        return error

    if isinstance(error, DependencyInstantiationError):
        if (<DependencyStack> stack).depth == 1:
            return DependencyInstantiationError(<object> dependency)
        else:
            return error
    return DependencyInstantiationError(<object> dependency,
                                        (<DependencyStack> stack).to_list()[1:])

#########
# CACHE #
#########

cdef struct Entry:
    PyObject *key
    CacheValue value

cdef Entry *create_table(size_t size):
    cdef:
        Entry *table
        Entry *entry

    table = <Entry *> PyMem_Malloc(size * sizeof(Entry))
    if not table:
        raise MemoryError()

    for entry in table[:size]:
        entry.key = NULL

    return table

cdef class DependencyCache:
    cdef:
        size_t mask
        size_t used
        Entry *table

    def __cinit__(self):
        cdef:
            size_t size = 8
        self.table = create_table(size)
        self.mask = size - 1
        self.used = 0

    def __dealloc__(self):
        cdef:
            Entry *entry
        for entry in self.table[:self.mask + 1]:
            if entry.key:
                Py_DECREF(entry.key)
                Py_DECREF(entry.value.ptr)
        PyMem_Free(self.table)

    # Python functions only used for tests.
    def __len__(self):
        return self.used

    def __setitem__(self, key, value):
        self.set(<PyObject *> key, 0, <PyObject *> value)

    def __getitem__(self, key):
        cdef:
            CacheValue *value = self.get(<PyObject *> key)
        if value is NULL:
            raise KeyError(key)
        return <object> value.ptr

    cdef CacheValue *get(self, PyObject *key):
        """ The Container may update the CacheValue """
        cdef:
            Entry *entry = self._find(key)
        if entry.key:
            return &entry.value
        return NULL

    cdef set(self, PyObject *key, Header header, PyObject *value):
        cdef:
            Entry *entry = self._find(key)
        if entry.key is NULL:
            self.used += 1
            Py_INCREF(key)
            Py_INCREF(value)
            entry.key = key
            entry.value.header = header
            entry.value.ptr = value

            if 3 * self.used > 2 * self.mask:
                self._resize()
        else:
            Py_INCREF(value)
            Py_DECREF(entry.value.ptr)
            entry.value.header = header
            entry.value.ptr = value

    cdef _resize(self):
        cdef:
            Entry *old_table = self.table
            size_t old_mask = self.mask
            size_t size = 8
            size_t minsize = 2 * self.used

        while size < minsize:
            size *= 2

        # We don't need thread-safety thanks to the GIL. Whenever we use set(),
        # it's because the value didn't exist and we don't keep any dangling
        # cached value.
        self.table = create_table(size)
        self.mask = size - 1
        self.used = 0

        cdef:
            Entry *old_entry
            Entry *entry
            size_t i
        for old_entry in old_table[:old_mask + 1]:
            if old_entry.key:
                entry = self._find(old_entry.key)
                entry[0] = old_entry[0]
                self.used += 1

        PyMem_Free(old_table)

    cdef Entry *_find(self, PyObject *key):
        cdef:
            size_t mask = self.mask
            Entry *table = self.table
            Entry *cursor
            size_t i, perturb
            size_t h = (<size_t> key)

        # Python objects are at least 2 * size_t long.
        h //= 2 * sizeof(size_t)
        # Mixing the entropy into the lower bytes. Same strategy than scala's OpenHashMap.
        i = h ^ (h >> 20) ^ (h >> 12)
        i ^= (i >> 7) ^ (i >> 4)

        # Using same strategy than Python dict
        perturb = h
        while True:
            cursor = &(table[i & mask])
            if cursor.key is key or cursor.key is NULL:
                return cursor
            perturb >>= 5
            i = (5 * i + 1) + perturb

############
# OVERRIDE #
############

cdef class OverridableRawContainer(RawContainer):
    cdef:
        object __override_lock
        dict __scopes_override
        dict __singletons_override
        dict __factory_overrides
        object __provider_overrides


    def __init__(self):
        super().__init__()
        self.__override_lock = threading.RLock()
        # Used to differentiate singletons from the overrides and the "normal" ones.
        self.__singletons_override = dict()
        self.__scopes_override = dict()  # type:  Dict[Scope, Dict[Hashable, object]]
        self.__factory_overrides = dict()  # type: Dict[Any, Tuple[Callable[[], Any], Optional[Scope]]]
        self.__provider_overrides = deque()  # type: Deque[Callable[[Any], Optional[DependencyValue]]]

    def clone(self,
              *,
              keep_singletons: bool = False,
              keep_scopes: bool = False) -> OverridableRawContainer:
        cdef:
            OverridableRawContainer clone

        with self.locked():
            clone = super().clone(keep_singletons=keep_singletons,
                                  keep_scopes=keep_scopes)
            if keep_singletons:
                clone.__singletons_override = self.__singletons_override
            clone.__scopes_override = {
                scope: dependencies.copy() if keep_scopes else dict()
                for scope, dependencies in self.__scopes_override.items()
            }
            clone.__factory_overrides = self.__factory_overrides
            clone.__provider_overrides = self.__provider_overrides

            return clone

    def override_singletons(self, singletons: dict):
        with self.__override_lock:
            self.__singletons_override.update(singletons)

    def override_factory(self,
                         dependency: Hashable,
                         *,
                         factory: Callable[[], Any],
                         scope: Optional[Scope]):
        with self.__override_lock:
            try:
                del self.__singletons_override[dependency]
            except KeyError:
                pass
            for scope_dependencies in self.__scopes_override.values():
                try:
                    del scope_dependencies[dependency]
                except KeyError:
                    pass
            self.__factory_overrides[dependency] = (factory, scope)

    def override_provider(self,
                          provider: Callable[[Any], Optional[DependencyValue]]):
        with self.__override_lock:
            self.__provider_overrides.appendleft(provider)  # latest provider wins

    def reset_scope(self, scope: Scope) -> None:
        super().reset_scope(scope)
        self.__scopes_override[scope] = dict()

    def clone(self,
              *,
              keep_singletons: bool,
              keep_scopes: bool) -> OverridableRawContainer:
        cdef:
            OverridableRawContainer container

        with self.__override_lock:
            container = super().clone(keep_singletons=keep_singletons,
                                      keep_scopes=keep_scopes)
            if keep_singletons:
                container.__singletons_override = self.__singletons_override.copy()
            container.__scopes_override = {
                scope: dependencies.copy() if keep_scopes else dict()
                for scope, dependencies in self.__scopes_override.items()
            }
            container.__factory_overrides = self.__factory_overrides.copy()
            container.__provider_overrides = self.__provider_overrides.copy()

        return container

    def debug(self, dependency: Hashable):
        from .._internal.utils.debug import debug_repr
        from .utils import DependencyDebug

        with self.__override_lock:
            try:
                value = self.__singletons_override[dependency]
            except KeyError:
                pass
            else:
                return DependencyDebug(f"Override/Singleton: {debug_repr(dependency)} "
                                       f"-> {value!r}",
                                       scope=Scope.singleton())
            try:
                (factory, scope) = self.__factory_overrides[dependency]
            except KeyError:
                pass
            else:
                return DependencyDebug(f"Override/Factory: {debug_repr(dependency)} "
                                       f"-> {debug_repr(factory)}",
                                       scope=scope)

        return super().debug(dependency)

    # Less efficient than the original fast_get, but we don't really care in tests.
    cdef fast_get(self, PyObject *dependency, DependencyResult *result):
        cdef:
            PyObject *ptr

        dep = <object> dependency
        result.value = NULL
        with self.__override_lock, self._instantiation_lock:
            with self._dependency_stack.instantiating(dep):
                try:
                    obj = self.__singletons_override[dep]
                except KeyError:
                    pass
                else:
                    DependencyValue(obj, scope=_SCOPE_SINGLETON).to_result(result)
                    return

                for scope, dependencies in self.__scopes_override.items():
                    try:
                        obj = dependencies[dep]
                    except KeyError:
                        pass
                    else:
                        DependencyValue(obj, scope=scope).to_result(result)
                        return

                try:
                    for provider in self.__provider_overrides:
                        value = provider(dep)
                        if value is not None:
                            if value.scope is Scope.singleton():
                                self.__singletons_override[dep] = value.unwrapped
                            elif value.scope is not None:
                                self.__scopes_override[value.scope][dep] = value.unwrapped
                            (<DependencyValue?> value).to_result(result)
                            return

                    try:
                        (factory, scope) = self.__factory_overrides[dep]
                    except KeyError:
                        pass
                    else:
                        value = factory()
                        if scope is Scope.singleton():
                            self.__singletons_override[dep] = value
                        elif scope is not None:
                            self.__scopes_override[scope][dep] = value
                        DependencyValue(value, scope=scope).to_result(result)
                        return

                except Exception as error:
                    new_error = handle_error(dependency,
                                             <PyObject *> self._dependency_stack,
                                             error)
                    if new_error is not error:
                        raise new_error from error
                    else:
                        raise

            RawContainer.fast_get(self, dependency, result)



from __future__ import annotations

import inspect
from typing import Callable, Dict, Hashable, Optional

from .._internal import API
from .._internal.utils import debug_repr, FinalImmutable
from ..core import Container, DependencyDebug, DependencyValue, Provider, Scope


@API.private
class IndirectProvider(Provider[Hashable]):
    def __init__(self) -> None:
        super().__init__()
        self.__implementations: Dict[ImplementationDependency, Hashable] = dict()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(" \
               f"implementations={list(self.__implementations.keys())})"

    def clone(self, keep_singletons_cache: bool) -> IndirectProvider:
        p = IndirectProvider()
        p.__implementations = self.__implementations.copy()
        return p

    def exists(self, dependency: Hashable) -> bool:
        return (isinstance(dependency, ImplementationDependency)
                and dependency in self.__implementations)

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        if not isinstance(dependency, ImplementationDependency):
            return None

        try:
            target = self.__implementations[dependency]
        except KeyError:
            return None

        if target is None:
            target = dependency.implementation()

        return DependencyDebug(debug_repr(dependency),
                               scope=Scope.singleton() if dependency.permanent else None,
                               wired=[dependency.implementation],  # type: ignore
                               dependencies=[target])

    def maybe_provide(self, dependency: Hashable, container: Container
                      ) -> Optional[DependencyValue]:
        if not isinstance(dependency, ImplementationDependency):
            return None

        try:
            target = self.__implementations[dependency]
        except KeyError:
            return None

        if target is not None:
            return container.provide(target)
        else:
            # Mypy treats linker as a method
            target = dependency.implementation()
            if dependency.permanent:
                self.__implementations[dependency] = target
            value = container.provide(target)
            return DependencyValue(
                value.unwrapped,
                scope=value.scope if dependency.permanent else None
            )

    def register_implementation(self,
                                interface: type,
                                implementation: Callable[[], Hashable],
                                *,
                                permanent: bool
                                ) -> ImplementationDependency:
        assert callable(implementation) \
               and inspect.isclass(interface) \
               and isinstance(permanent, bool)
        impl = ImplementationDependency(interface, implementation, permanent)
        self._assert_not_duplicate(impl)
        self.__implementations[impl] = None
        return impl


@API.private
class ImplementationDependency(FinalImmutable):
    __slots__ = ('interface', 'implementation', 'permanent', '__hash')
    interface: type
    implementation: Callable[[], Hashable]
    permanent: bool
    __hash: int

    def __init__(self,
                 interface: object,
                 implementation: Callable[[], Hashable],
                 permanent: bool):
        super().__init__(interface,
                         implementation,
                         permanent,
                         hash((interface, implementation)))

    def __repr__(self) -> str:
        return f"Implementation({self})"

    def __antidote_debug_repr__(self) -> str:
        if self.permanent:
            return f"Permanent implementation: {self}"
        else:
            return f"Implementation: {self}"

    def __str__(self) -> str:
        impl = self.implementation  # type: ignore
        return f"{debug_repr(self.interface)} @ {debug_repr(impl)}"

    # Custom hash & eq necessary to find duplicates
    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, ImplementationDependency)
                and self.__hash == other.__hash
                and (self.interface is other.interface
                     or self.interface == other.interface)
                and (self.implementation is other.implementation  # type: ignore
                     or self.implementation == other.implementation)  # type: ignore
                )  # noqa



from __future__ import annotations

import collections.abc as c_abc
import inspect
import warnings
from typing import (Any, Callable, Generic, Hashable, Iterable, Mapping, Optional,
                    overload,
                    Sequence, Type, TYPE_CHECKING, TypeVar, Union)

from typing_extensions import final, TypeAlias

from .annotations import Get
from .marker import InjectMeMarker
from .._internal import API
from .._internal.utils import FinalImmutable
from .._internal.utils.meta import Singleton

if TYPE_CHECKING:
    from .typing import CallableClass, Source

F = TypeVar('F', bound=Callable[..., Any])
T = TypeVar('T')
AnyF: TypeAlias = 'Union[Callable[..., Any], staticmethod[Any], classmethod[Any]]'


@API.deprecated
@final
class Arg(FinalImmutable):
    """
    .. deprecated:: 1.1

        Deprecated as specifying a function to :py:func:`~.injection.inject` is deprecated

    Represents an argument (name and type hint) if you need a very custom injection
    logic.
    """
    __slots__ = ('name', 'type_hint', 'type_hint_with_extras')
    name: str
    """Name of the argument"""
    type_hint: Any
    """Type hint of the argument if any"""
    type_hint_with_extras: Any
    """
    Type hint of the argument if any with include_extras=True, so with annotations.
    """

    def __init__(self, name: str, type_hint: Any, type_hint_with_extras: Any) -> None:
        warnings.warn("Deprecated as specifying a function to @inject is deprecated",
                      DeprecationWarning)
        super().__init__(name, type_hint, type_hint_with_extras)


# API.experimental
DEPENDENCIES_TYPE: TypeAlias = Union[
    None,
    Mapping[str, Hashable],  # {arg_name: dependency, ...}
    Sequence[Optional[Hashable]],  # (dependency for arg 1, ...)
    # API.Deprecated
    Callable[[Arg], Optional[Hashable]],  # arg -> dependency
]
# API.deprecated
AUTO_PROVIDE_TYPE: TypeAlias = Optional[Union[
    bool,  # all class type hints or nothing
    Iterable[type],  # specific list of classes
    Callable[[type], bool]  # Function determining which classes should be auto provided
]]


class InjectedCallableMeta(type):
    def __instancecheck__(self, instance: object) -> bool:
        from .._internal.wrapper import is_wrapper
        return is_wrapper(instance)


class InjectedCallable(Generic[F], metaclass=InjectedCallableMeta):
    __wrapped__: F

    def __call__(self, *args: object, **kwargs: object) -> object:
        ...  # pragma: no cover


@API.private  # Use the singleton instance `inject`, not the class directly.
class Inject(Singleton):
    """
    Use :py:obj:`.inject` directly, this class is not meant to instantiated or
    subclassed.
    """

    @API.public
    def me(self,
           *,
           source: Optional[Union[
               Source[Any],
               Callable[..., Any],
               Type[CallableClass[Any]]
           ]] = None
           ) -> Any:
        """
        Injection Marker specifying that the current type hint should be used as dependency.

        .. doctest:: core_inject_me

            >>> from antidote import inject, service
            >>> @service
            ... class MyService:
            ...     pass
            >>> @inject
            ... def f(s: MyService = inject.me()) -> MyService:
            ...     return s
            >>> f()
            <MyService object at ...>

        It also works with dependency sources such as :py:func:`~.factory.factory` or
        :py:func:`.implementation`:

        .. doctest:: core_inject_me_from

            >>> from antidote import inject, factory
            >>> class Dummy:
            ...     pass
            >>> @factory
            ... def current_dummy() -> Dummy:
            ...     return Dummy()
            >>> @inject
            ... def f(d: Dummy = inject.me(source=current_dummy)) -> Dummy:
            ...     return d
            >>> f()
            <Dummy object at ...>

        When the type hint is :code:`Optional` :py:func:`.inject` won't raise
        :py:exc:`~.exceptions.DependencyNotFoundError` but will provide :py:obj:`None` instead:

        .. doctest:: core_inject_me_optional

            >>> from typing import Optional
            >>> from antidote import inject
            >>> class MyService:
            ...     pass
            >>> @inject
            ... def f(s: Optional[MyService] = inject.me()) -> MyService:
            ...     return s
            >>> f() is None
            True

        """
        return InjectMeMarker(source=source)

    @overload
    def get(self, __dependency: object) -> Any:
        ...  # pragma: no cover

    @overload
    def get(self,
            __dependency: Type[T],
            *,
            source: Union[Source[T], Callable[..., T], Type[CallableClass[T]]]
            ) -> Any:
        ...  # pragma: no cover

    @API.public
    def get(self,
            __dependency: Any,
            *,
            source: Optional[Union[
                Source[Any],
                Callable[..., Any],
                Type[CallableClass[Any]]
            ]] = None
            ) -> Any:
        """
        Injection Marker specifying explicitly which dependency to retrieve.

        .. doctest:: core_inject_get

            >>> from antidote import inject, service
            >>> @service
            ... class MyService:
            ...     pass
            >>> @inject
            ... def f(s = inject.get(MyService)):
            ...     return s
            >>> f()
            <MyService object at ...>

        """
        return Get(__dependency, source=source) if source is not None else Get(__dependency)

    @overload
    def __call__(self,
                 __arg: staticmethod[F],
                 *,
                 dependencies: DEPENDENCIES_TYPE = None,
                 auto_provide: API.Deprecated[AUTO_PROVIDE_TYPE] = None,
                 strict_validation: bool = True,
                 ignore_type_hints: bool = False
                 ) -> staticmethod[F]:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 __arg: classmethod[F],
                 *,
                 dependencies: DEPENDENCIES_TYPE = None,
                 auto_provide: API.Deprecated[AUTO_PROVIDE_TYPE] = None,
                 strict_validation: bool = True,
                 ignore_type_hints: bool = False
                 ) -> classmethod[F]:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 __arg: F,
                 *,
                 dependencies: DEPENDENCIES_TYPE = None,
                 auto_provide: API.Deprecated[AUTO_PROVIDE_TYPE] = None,
                 strict_validation: bool = True,
                 ignore_type_hints: bool = False
                 ) -> F:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 *,
                 dependencies: DEPENDENCIES_TYPE = None,
                 auto_provide: API.Deprecated[AUTO_PROVIDE_TYPE] = None,
                 strict_validation: bool = True,
                 ignore_type_hints: bool = False
                 ) -> Callable[[F], F]:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 __arg: Sequence[object],
                 *,
                 auto_provide: API.Deprecated[AUTO_PROVIDE_TYPE] = None,
                 strict_validation: bool = True,
                 ignore_type_hints: bool = False
                 ) -> Callable[[F], F]:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 __arg: Mapping[str, object],
                 *,
                 auto_provide: API.Deprecated[AUTO_PROVIDE_TYPE] = None,
                 strict_validation: bool = True,
                 ignore_type_hints: bool = False
                 ) -> Callable[[F], F]:
        ...  # pragma: no cover

    @API.public
    def __call__(self,
                 __arg: Union[None, AnyF, Sequence[object], Mapping[str, object]] = None,
                 *,
                 dependencies: DEPENDENCIES_TYPE = None,
                 auto_provide: API.Deprecated[AUTO_PROVIDE_TYPE] = None,
                 strict_validation: bool = True,
                 ignore_type_hints: bool = False
                 ) -> AnyF:
        """
        Inject the dependencies into the function lazily, they are only retrieved
        upon execution. As several options can apply to the same argument, the priority is
        defined as:

        1. Markers :code:`inject.me`... / Annotated type hints (PEP-593)
        2. Dependencies declared explicitly if any declared with :code:`dependencies`.
        3. Deprecated: Class type hints if specified with :code:`auto_provide`.

        .. doctest:: core_inject

            >>> from antidote import inject, service, Inject
            >>> @service
            ... class A:
            ...     pass
            >>> @service
            ... class B:
            ...     pass
            >>> @inject
            ... def f(a: A = inject.me()):
            ...     pass # a = world.get(A)
            >>> # PEP-593 annotation
            ... @inject
            ... def f(a: Inject[A]):
            ...     pass # a = world.get(A)
            >>> # All possibilities for dependency argument.
            >>> @inject(dict(b='dependency'))  # shortcut
            ... def f(a, b):
            ...     pass
            >>> @inject([None, 'dependency'])  # shortcut
            ... def f(a, b):
            ...     pass  # a, b = <not injected>, world.get('dependency')

        Args:
            ignore_type_hints:
            __arg: Callable to be wrapped. Can also be used on static methods or class
                methods. May also be sequence of dependencies or mapping from argument
                name to dependencies.
            dependencies:
                Explicit definition of the dependencies which overrides :code:`auto_provide`.
                Defaults to :py:obj:`None`. Can be one of:

                - Mapping from argument name to its dependency
                - Sequence of dependencies which will be mapped with the position
                  of the arguments. :py:obj:`None` can be used as a placeholder.

                .. deprecated:: 1.1
                    Specifying a callable is deprecated. The Callable receives
                    :py:class:`~.Arg` as arguments and should return the matching
                    dependency. :py:obj:`None` should be used for arguments without dependency.

            auto_provide:
                .. deprecated:: 1.1

                Whether or not class type hints should be used as the arguments
                dependency. Only classes are taken into account and it works with
                :py:class:`~typing.Optional`. An iterable of classes  may also be supplied
                to activate this feature only for those. A function may also be provided, in
                which case it'll be called to determine whether the class type hint ca be
                provided or not. Defaults to :code:`False`.
            strict_validation: Whether arguments should be strictly validated given the
                decorated function's argumnet. For example, a key in the dependencies dict
                that does not match any argument would raise error. Defaults to
                :py:obj:`True`.

        Returns:
            The decorator to be applied or the injected function if the
            argument :code:`func` was supplied.

        """
        from ._injection import raw_inject

        if isinstance(__arg, (c_abc.Sequence, c_abc.Mapping)):
            if isinstance(__arg, str):
                raise TypeError("First argument must be an sequence/mapping of dependencies "
                                "or the function to be wrapped, not a string.")
            if dependencies is not None:
                raise TypeError("dependencies must be None if a sequence/mapping of "
                                "dependencies is provided as first argument.")
            dependencies = __arg
            __arg = None

        if callable(dependencies):
            warnings.warn("""
            Passing a callable to dependencies is deprecated.
            If you rely on this behavior, wrap @inject instead.
            """, DeprecationWarning)

        if auto_provide is not None:
            warnings.warn("""
            Using auto_provide is deprecated.
            If you rely on this behavior, wrap @inject instead.
            """, DeprecationWarning)

        def decorate(f: AnyF) -> AnyF:
            return raw_inject(
                f,
                dependencies=dependencies,
                auto_provide=auto_provide if auto_provide is not None else False,
                strict_validation=strict_validation,
                ignore_type_hints=ignore_type_hints
            )

        return __arg and decorate(__arg) or decorate


inject = Inject()
inject.__doc__ = \
    """
    Singleton instance of :py:class:`~.core.injection.Inject`
    """


@API.public  # Function will be kept in sync with @inject, so you may use it.
def validate_injection(dependencies: DEPENDENCIES_TYPE = None,
                       auto_provide: API.Deprecated[AUTO_PROVIDE_TYPE] = None) -> None:
    """
    Validates that injection parameters are valid. If not, an error is raised.

    .. doctest:: core_injection_validate_injection

        >>> from antidote.utils import validate_injection
        >>> validate_injection(dependencies={'name': object()})
        >>> validate_injection(dependencies=object())
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        TypeError: ...


    Args:
        dependencies: As defined by :py:func:`~.injection.inject`
        auto_provide:
            .. deprecated:: 1.1

            As defined by :py:func:`~.injection.inject`
    """

    if auto_provide is not None:
        warnings.warn("Using auto_provide is deprecated.", DeprecationWarning)

    if not (dependencies is None
            or (isinstance(dependencies, (c_abc.Sequence, c_abc.Mapping))
                and not isinstance(dependencies, str))
            or callable(dependencies)):
        raise TypeError(
            f"dependencies can be None, a mapping of arguments names to dependencies, "
            f"a sequence of dependencies, a function, "
            f"not a {type(dependencies)!r}"
        )
    if isinstance(dependencies, c_abc.Mapping):
        if not all(isinstance(k, str) for k in dependencies.keys()):
            raise TypeError("Dependencies keys must be argument names (str)")

    if isinstance(auto_provide, str) \
            or not (auto_provide is None
                    or callable(auto_provide)
                    or isinstance(auto_provide, (bool, c_abc.Iterable))):
        raise TypeError(
            f"auto_provide must be either a boolean or an iterable of classes, "
            f"not {type(auto_provide)!r}.")

    # If we can iterate over it safely
    if isinstance(auto_provide, (list, set, tuple, frozenset)):
        for cls in auto_provide:
            if not (isinstance(cls, type) and inspect.isclass(cls)):
                raise TypeError(f"auto_provide must be a boolean or an iterable of "
                                f"classes, but contains {cls!r} which is not a class.")



from contextlib import contextmanager
from typing import Any, Callable

import pytest

from antidote import Factory, factory, inject, Inject, Provide, Service, Wiring, world
from antidote._providers import (FactoryProvider, LazyProvider, ServiceProvider)
from antidote.exceptions import DependencyInstantiationError


@contextmanager
def does_not_raise():
    yield


@pytest.fixture(autouse=True)
def test_world():
    with world.test.empty():
        world.provider(ServiceProvider)
        world.provider(FactoryProvider)
        world.provider(LazyProvider)
        yield


class A:
    def __init__(self, **kwargs):
        self.kwargs = kwargs


class B:
    pass


class C:
    pass


def dummy() -> A:
    pass


@pytest.fixture(params=['function', 'class', 'decorated_class'])
def build(test_world, request):
    if request.param == 'function':
        @factory
        def build(**kwargs) -> A:
            return A(**kwargs)

        return build
    elif request.param == "class":
        class ServiceFactory(Factory):
            def __call__(self, **kwargs) -> A:
                return A(**kwargs)

        return ServiceFactory
    else:
        @factory
        class ServiceFactory:
            def __call__(self, **kwargs) -> A:
                return A(**kwargs)

        return ServiceFactory


def test_simple(build: Callable[..., A]):
    # working dependency
    assert isinstance(world.get(A, source=build), A)
    assert world.get(A, source=build) is world.get(A, source=build)


def test_pass_through(build):
    if isinstance(build, type):
        build = build()

    assert isinstance(build(), A)

    x = object()
    a = build(x=x)
    assert a.kwargs == dict(x=x)


def test_legacy_notation():
    @factory
    def f() -> A:
        return A()

    class BFactory(Factory):
        def __call__(self) -> B:
            return B()

    assert isinstance(world.get(A @ f), A)
    assert isinstance(world.get(B @ BFactory), B)


def test_custom_scope():
    dummy_scope = world.scopes.new(name='dummy')

    class Scoped:
        pass

    class ScopedF(Factory):
        __antidote__ = Factory.Conf(scope=dummy_scope)

        def __call__(self) -> Scoped:
            return Scoped()

    x = world.get(Scoped @ ScopedF)
    assert world.get(Scoped @ ScopedF) is x
    world.scopes.reset(dummy_scope)
    assert world.get(Scoped @ ScopedF) is not x

    @factory(scope=dummy_scope)
    def scoped_factory() -> Scoped:
        return Scoped()

    x = world.get(Scoped @ scoped_factory)
    assert world.get(Scoped @ scoped_factory) is x
    world.scopes.reset(dummy_scope)
    assert world.get(Scoped @ scoped_factory) is not x


def test_parameterized():
    x = object()

    class BuildA(Factory):
        __antidote__ = Factory.Conf(parameters=['x'])

        def __call__(self, **kwargs) -> A:
            return A(**kwargs)

    a = world.get(A @ BuildA.parameterized(x=x))
    assert a.kwargs == dict(x=x)

    with pytest.raises(ValueError, match=".*parameters.*'x'.*"):
        A @ BuildA.parameterized()

    with pytest.raises(ValueError, match=".*parameters.*'x'.*"):
        A @ BuildA.parameterized(unknown='something')


def test_not_parametrized():
    class BuildA(Factory):
        def __call__(self) -> A:
            return A()

    with pytest.raises(RuntimeError, match=".*parameters.*"):
        B @ BuildA.parameterized(x=1)


def test_invalid_with_default_parameters():
    with pytest.raises(ValueError, match=".*default.*"):
        class BuildA(Factory):
            __antidote__ = Factory.Conf(parameters=['x'])

            def __call__(self, x: str = 'default') -> A:
                return A()


def test_invalid_with_injected_parameters():
    class Dummy(Service):
        pass

    with pytest.raises(ValueError, match=".*injected.*class.*Dummy.*"):
        class BuildA(Factory):
            __antidote__ = Factory.Conf(parameters=['x'])

            def __call__(self, x: Provide[Dummy]) -> A:
                return A()


def test_invalid_parameterized_dependency():
    class BuildA(Factory):
        __antidote__ = Factory.Conf(parameters=['x'])

        def __call__(self, x) -> A:
            return A()

    with pytest.raises(ValueError, match="Unsupported output.*"):
        B @ BuildA.parameterized(x=1)


def test_invalid_dependency():
    @factory
    def f() -> A:
        pass

    with pytest.raises(ValueError, match="Unsupported output.*"):
        B @ f

    class F(Factory):
        def __call__(self) -> A:
            pass

    with pytest.raises(ValueError, match="Unsupported output.*"):
        B @ F


def test_getattr():
    def build() -> A:
        return A()

    build.hello = 'world'

    build = factory(build)
    assert build.hello == 'world'

    build.new_hello = 'new_world'
    assert build.new_hello == 'new_world'


def test_wiring():
    world.test.singleton(A, A())

    @factory
    def build_b(a: A = inject.me()) -> B:
        return B()

    @factory
    class BuildB:
        def __call__(self, a: A = inject.me()) -> B:
            return B()

    assert isinstance(world.get(B, source=build_b), B)
    assert isinstance(world.get(B, source=BuildB), B)


def test_wiring_none():
    world.test.singleton(A, A())

    @factory(wiring=None)
    def build_b2(a: Inject[A]) -> B:
        return B()

    @factory(wiring=None)
    class BuildB2:
        def __call__(self, a: Inject[A]) -> B:
            return B()

    with pytest.raises(DependencyInstantiationError):
        world.get(B, source=build_b2)

    with pytest.raises(TypeError):
        build_b2()

    with pytest.raises(DependencyInstantiationError):
        world.get(B, source=BuildB2)

    with pytest.raises(TypeError):
        BuildB2()()


def test_wiring_custom():
    world.test.singleton(B, B())

    @factory(wiring=Wiring(dependencies=dict(b=B)))
    def build_b3(b) -> B:
        return b

    @factory(wiring=Wiring(dependencies=dict(b=B)))
    class BuildB3:
        def __call__(self, b) -> B:
            return b

    assert world.get(B, source=build_b3) is world.get(B)
    assert world.get(B, source=BuildB3) is world.get(B)


def test_missing_call():
    with pytest.raises(TypeError, match="__call__"):
        class ServiceFactory(Factory):
            pass

    with pytest.raises(ValueError, match=".*return type hint.*"):
        @factory
        class ServiceFactory2:
            pass


def test_missing_return_type_hint():
    with pytest.raises(ValueError, match=".*return type hint.*"):
        @factory
        def faulty_service_provider():
            return A()

    with pytest.raises(TypeError, match=".*return type hint.*"):
        @factory
        def faulty_service_provider2() -> Any:
            return A()

    with pytest.raises(ValueError, match=".*return type hint.*"):
        @factory
        class FaultyServiceFactory:
            def __call__(self):
                return A()

    with pytest.raises(TypeError, match=".*return type hint.*"):
        @factory
        class FaultyServiceFactory2:
            def __call__(self) -> Any:
                return A()

    with pytest.raises(ValueError, match=".*return type hint.*"):
        class FaultyServiceFactory3(Factory):
            def __call__(self):
                return A()

    with pytest.raises(TypeError, match=".*return type hint.*"):
        class FaultyServiceFactory4(Factory):
            def __call__(self) -> Any:
                return A()


@pytest.mark.parametrize('expectation,kwargs,func',
                         [
                             pytest.param(pytest.raises(TypeError, match='.*function.*'),
                                          dict(),
                                          object(),
                                          id='function'),
                             pytest.param(pytest.raises(TypeError, match='.*singleton.*'),
                                          dict(singleton=object()),
                                          lambda: None,
                                          id='singleton'),
                             pytest.param(pytest.raises(TypeError, match='.*scope.*'),
                                          dict(scope=object()),
                                          lambda: None,
                                          id='scope'),
                             pytest.param(pytest.raises(TypeError, match='.*wiring.*'),
                                          dict(wiring=object()),
                                          lambda: None,
                                          id='wiring'),
                         ])
def test_invalid_factory_args(expectation, kwargs: dict, func: Callable[..., object]):
    with expectation:
        factory(**kwargs)(func)

    with expectation:
        factory(func, **kwargs)


def test_no_subclass_of_service():
    class Dummy(Factory):
        def __call__(self, *args, **kwargs) -> A:
            pass

    with pytest.raises(TypeError, match=".*abstract.*"):
        class SubDummy(Dummy):
            def __call__(self, *args, **kwargs) -> B:
                pass


def test_invalid_conf():
    with pytest.raises(TypeError, match=".*__antidote__.*"):
        class Dummy(Factory):
            __antidote__ = object()

            def __call__(self) -> A:
                pass


@pytest.mark.parametrize('expectation, parameters', [
    (pytest.raises(TypeError), "string"),
    (pytest.raises(TypeError), object()),
    (pytest.raises(TypeError), [1]),
    (does_not_raise(), ['x']),
    (does_not_raise(), []),
    (does_not_raise(), None),
])
def test_conf_parameters(expectation, parameters):
    with expectation:
        class Build(Factory):
            __antidote__ = Factory.Conf(parameters=parameters)

            def __call__(self, *args, **kwargs) -> A:
                pass


@pytest.mark.parametrize('expectation, kwargs', [
    pytest.param(pytest.raises(TypeError, match=f'.*{arg}.*'),
                 {arg: object()},
                 id=arg)
    for arg in ['wiring',
                'singleton',
                'scope',
                'parameters']
])
def test_invalid_conf_args(kwargs, expectation):
    with expectation:
        Factory.Conf(**kwargs)


@pytest.mark.parametrize('kwargs', [
    dict(singleton=False),
    dict(scope=None),
    dict(wiring=Wiring(methods=['method'])),
    dict(parameters=frozenset(['x']))
])
def test_conf_copy(kwargs):
    conf = Factory.Conf(singleton=True).copy(**kwargs)
    for k, v in kwargs.items():
        assert getattr(conf, k) == v


def test_invalid_copy():
    conf = Factory.Conf()
    with pytest.raises(TypeError, match=".*both.*"):
        conf.copy(singleton=False, scope=None)


def test_conf_repr():
    conf = Factory.Conf()
    assert "scope" in repr(conf)


def test_default_injection():
    class MyService(Service):
        pass

    class A:
        pass

    injected = None

    @factory
    def build_a(s: Provide[MyService]) -> A:
        nonlocal injected
        injected = s
        return A()

    assert isinstance(world.get(A @ build_a), A)
    assert injected is world.get(MyService)


def test_double_injection():
    world.test.singleton(B, object())

    injected = None

    @factory
    @inject(auto_provide=True)
    def build_a(s: B) -> A:
        nonlocal injected
        injected = s
        return A()

    assert isinstance(world.get(A @ build_a), A)
    assert injected is world.get(B)



from __future__ import annotations

import inspect
from typing import cast, Dict, Hashable, Optional

from .._internal import API
from .._internal.utils import debug_repr, FinalImmutable
from ..core import Container, DependencyDebug, DependencyValue, Provider, Scope


@API.private
class Parameterized(FinalImmutable):
    __slots__ = ('wrapped', 'parameters', '_hash')
    wrapped: Hashable
    parameters: Dict[str, object]
    _hash: int

    def __init__(self, dependency: Hashable, parameters: Dict[str, object]) -> None:
        assert isinstance(parameters, dict) and parameters

        try:
            # Try most precise hash first
            _hash = hash((dependency, tuple(parameters.items())))
        except TypeError:
            # If type error, use the best error-free hash possible
            _hash = hash((dependency, tuple(parameters.keys())))

        super().__init__(dependency, parameters, _hash)

    def __hash__(self) -> int:
        return self._hash

    def __repr__(self) -> str:
        return f"Parameterized(dependency={self.wrapped}, parameters={self.parameters})"

    def __antidote_debug_repr__(self) -> str:
        return f"{debug_repr(self.wrapped)} with parameters={self.parameters}"

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, Parameterized)
                and self._hash == other._hash
                and (self.wrapped is other.wrapped
                     or self.wrapped == other.wrapped)
                and self.parameters == other.parameters)  # noqa


@API.private
class ServiceProvider(Provider[Hashable]):
    def __init__(self) -> None:
        super().__init__()
        self.__services: Dict[object, Optional[Scope]] = dict()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(services={list(self.__services.items())!r})"

    def exists(self, dependency: Hashable) -> bool:
        if isinstance(dependency, Parameterized):
            return dependency.wrapped in self.__services
        return dependency in self.__services

    def clone(self, keep_singletons_cache: bool) -> ServiceProvider:
        p = ServiceProvider()
        p.__services = self.__services.copy()
        return p

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        klass = dependency.wrapped if isinstance(dependency, Parameterized) else dependency
        try:
            scope = self.__services[klass]
        except KeyError:
            return None
        return DependencyDebug(debug_repr(dependency),
                               scope=scope,
                               wired=[klass])

    def maybe_provide(self, dependency: Hashable, container: Container
                      ) -> Optional[DependencyValue]:
        dep = dependency.wrapped if isinstance(dependency, Parameterized) else dependency
        try:
            scope = self.__services[dep]
        except KeyError:
            return None

        klass = cast(type, dep)
        if isinstance(dependency, Parameterized):
            instance = klass(**dependency.parameters)
        else:
            instance = klass()

        return DependencyValue(instance, scope=scope)

    def register(self,
                 klass: type,
                 *,
                 scope: Optional[Scope]
                 ) -> None:
        assert inspect.isclass(klass) \
               and (isinstance(scope, Scope) or scope is None)
        self._assert_not_duplicate(klass)
        self.__services[klass] = scope



from contextlib import contextmanager

import pytest
from typing_extensions import Annotated, Protocol, runtime_checkable

from antidote._internal.utils import enforce_type_if_possible


@contextmanager
def does_not_raise():
    yield


does_raise = pytest.raises(TypeError, match="(?i).*Value.*")


class DummyProtocol(Protocol):
    def dummy(self) -> None:
        pass


@runtime_checkable
class DummyRuntimeProtocol(Protocol):
    def dummy(self) -> None:
        pass


class ValidDummy:
    def dummy(self) -> None:
        pass


class InvalidDummy:
    pass


class SubDummy(ValidDummy):
    pass


@pytest.mark.parametrize('expectation, obj, tpe', [
    (does_raise, object(), int),
    (does_not_raise(), object(), Annotated[ValidDummy, object()]),
    (does_not_raise(), 1, int),
    (does_not_raise(), 1, DummyProtocol),
    (does_raise, 1, DummyRuntimeProtocol),
    (does_not_raise(), InvalidDummy(), DummyProtocol),
    (does_raise, InvalidDummy(), DummyRuntimeProtocol),
    (does_not_raise(), ValidDummy(), DummyProtocol),
    (does_not_raise(), ValidDummy(), DummyRuntimeProtocol),
    (does_not_raise(), SubDummy(), DummyProtocol),
    (does_not_raise(), SubDummy(), DummyRuntimeProtocol),
    (does_not_raise(), SubDummy(), ValidDummy),
    (does_raise, InvalidDummy(), ValidDummy),
])
def test_enforce_type(expectation, obj, tpe):
    with expectation:
        enforce_type_if_possible(obj, tpe)



from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import Callable, Dict, Hashable, List, Optional

from .service import Parameterized
from .._internal import API
from .._internal.utils import debug_repr, FinalImmutable
from ..core import (Container, DependencyDebug, DependencyValue, Provider,
                    Scope)


@API.private
class FactoryProvider(Provider[Hashable]):
    def __init__(self) -> None:
        super().__init__()
        self.__factories: Dict[FactoryDependency, Factory] = dict()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(factories={list(self.__factories.keys())})"

    def clone(self, keep_singletons_cache: bool) -> FactoryProvider:
        p = FactoryProvider()
        if keep_singletons_cache:
            factories = {
                k: (f.copy() if f.dependency is not None else f)
                for k, f in self.__factories.items()
            }
        else:
            factories = {
                k: (f.copy(keep_function=False) if f.dependency is not None else f)
                for k, f in self.__factories.items()
            }
        p.__factories = factories
        return p

    def exists(self, dependency: Hashable) -> bool:
        # For now we don't support multiple factories for a single dependency. Neither
        # is sharing the dependency with another provider. Simply because I don't see a
        # use case where it would make sense.
        # Open for discussions though, create an issue if you a use case.
        if isinstance(dependency, Parameterized):
            dependency = dependency.wrapped
        return (isinstance(dependency, FactoryDependency)
                and dependency in self.__factories)

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        dependency_factory = (dependency.wrapped
                              if isinstance(dependency, Parameterized)
                              else dependency)
        if not isinstance(dependency_factory, FactoryDependency):
            return None

        try:
            factory = self.__factories[dependency_factory]
        except KeyError:
            return None

        dependencies: List[object] = []
        wired: List[Callable[..., object]] = []
        if factory.function is None:
            dependencies.append(factory.dependency)
            if isinstance(factory.dependency, type) \
                    and inspect.isclass(factory.dependency):
                wired.append(factory.dependency.__call__)
        else:
            wired.append(factory.function)

        return DependencyDebug(debug_repr(dependency),
                               scope=factory.scope,
                               wired=wired,
                               dependencies=dependencies)

    def maybe_provide(self, dependency: Hashable, container: Container
                      ) -> Optional[DependencyValue]:
        dependency_factory = (dependency.wrapped
                              if isinstance(dependency, Parameterized)
                              else dependency)
        if not isinstance(dependency_factory, FactoryDependency):
            return None

        try:
            factory = self.__factories[dependency_factory]
        except KeyError:
            return None

        if factory.function is None:
            f = container.provide(factory.dependency)
            assert f.is_singleton(), "factory dependency is expected to be a singleton"
            assert callable(f.unwrapped)
            factory.function = f.unwrapped

        if isinstance(dependency, Parameterized):
            instance: object = factory.function(**dependency.parameters)
        else:
            instance = factory.function()

        return DependencyValue(instance, scope=factory.scope)

    def register(self,
                 output: type,
                 *,
                 scope: Optional[Scope],
                 factory: Optional[Callable[..., object]] = None,
                 factory_dependency: Optional[object] = None
                 ) -> FactoryDependency:
        assert inspect.isclass(output) \
               and (factory is None or factory_dependency is None) \
               and (factory is None or callable(factory)) \
               and (isinstance(scope, Scope) or scope is None)

        dependency = FactoryDependency(
            output=output,
            factory=factory or factory_dependency
        )
        self._assert_not_duplicate(dependency)

        if factory_dependency:
            self.__factories[dependency] = Factory(scope=scope,
                                                   function=None,
                                                   dependency=factory_dependency)
        else:
            self.__factories[dependency] = Factory(scope=scope,
                                                   function=factory,
                                                   dependency=None)

        return dependency


@API.private
class FactoryDependency(FinalImmutable):
    __slots__ = ('output', 'factory', '__hash')
    output: object
    factory: object
    __hash: int

    def __init__(self, *, output: object, factory: object):
        super().__init__(output, factory, hash((output, factory)))

    def __repr__(self) -> str:
        return f"FactoryDependency({self})"

    def __antidote_debug_repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f"{debug_repr(self.output)} @ {debug_repr(self.factory)}"

    # Custom hash & eq necessary to find duplicates
    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, FactoryDependency)
                and self.__hash == other.__hash
                and (self.output is other.output
                     or self.output == other.output)
                and (self.factory is other.factory
                     or self.factory == other.factory))  # noqa


@API.private
@dataclass
class Factory:
    __slots__ = ('scope', 'function', 'dependency')
    scope: Optional[Scope]
    dependency: Optional[Hashable]
    function: Optional[Callable[..., object]]

    def __post_init__(self) -> None:
        assert self.function is not None or self.dependency is not None

    def copy(self, keep_function: bool = True) -> Factory:
        return Factory(scope=self.scope,
                       function=self.function if keep_function else None,
                       dependency=self.dependency)



# @formatter:off
from typing import Hashable

cimport cython
from cpython.object cimport PyObject

from antidote.core.container cimport (Container, DependencyResult, DependencyValue, FastProvider)
from .._internal.utils import debug_repr
from ..core import DependencyDebug
from ..core.exceptions import DebugNotAvailableError

# @formatter:on

cdef extern from "Python.h":
    int PyObject_IsInstance(PyObject *inst, PyObject *cls) except -1


cdef class Lazy:
    def __antidote_debug_info__(self) -> DependencyDebug:
        raise DebugNotAvailableError()  # pragma: no cover

    cpdef __antidote_provide__(self, container: Container):
        raise NotImplementedError()  # pragma: no cover

    cdef fast_provide(self, PyObject*container, DependencyResult*result):
        cdef:
            DependencyValue dependency_instance
        dependency_instance = self.__antidote_provide__(<Container> container)
        if dependency_instance is not None:
            dependency_instance.to_result(result)

@cython.final
cdef class LazyProvider(FastProvider):
    cpdef LazyProvider clone(self, bint keep_singletons_cache):
        return LazyProvider.__new__(LazyProvider)

    def exists(self, dependency: Hashable) -> bool:
        return isinstance(dependency, Lazy)

    def maybe_debug(self, dependency: Hashable):
        if isinstance(dependency, Lazy):
            try:
                return dependency.__antidote_debug_info__()
            except DebugNotAvailableError:
                import warnings
                warnings.warn(f"Debug information for {debug_repr(dependency)} "
                              f"not available in {type(self)}")

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        if PyObject_IsInstance(dependency, <PyObject*> Lazy):
            return (<Lazy> dependency).fast_provide(container, result)
